#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_20(CoAPTestcase):
    """Identifier:
TD_COAP_CORE_20
Objective:
Perform GET transaction containing the Accept option (CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] clause 5.8.1,5.10.5,5.10.4

Pre-test
conditions:
•	Server should provide a resource /multi-format which exists in two formats:
-	text/plain;charset=utf-8
-	application/xml

Test Sequence:
Step
Type
Description

Part A: client requests a resource in text format


1
Stimulus
Client is requested to send a confirmable GET request to
server’s resource


2
Check
The request sent request by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Option type = Accept, value = 0 (text/plain;charset=utf-8)
•	Uri-Path option "multi-format"

3
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Message ID = CMID, Token = CTOK
•	Option type = Content-Format, value = 0 (text/plain;charset=utf-8)
•	Payload = Content of the requested resource in text/plain;charset=utf-8 format

4
Verify
Client displays the response

Part B: client requests a resource in xml format

5
Stimulus
Client is requested to send a confirmable GET request to
server’s resource


6
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Another client-generated Message ID ≠ CMID (➔ CMID2)
•	Client-generated Token which may or may not be ≠ CTOK (➔ CTOK2)
•	Option type = Accept, value = 41 (application/xml)

7
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Message ID = CMID2, Token = CTOK2
•	Option type = Content-Format, value = 41 (application/xml)
Payload = Content of the requested resource in application/xml format

8
Verify
Client displays the response
"""

    def run(self):
        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=self.uri("/multi-format") if self.urifilter else Opt(CoAPOptionAccept())))
        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=Opt(CoAPOptionAccept(0))))
        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        self.next_skip_ack()

        self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                       code=2.05,
                                       mid=CMID,
                                       tok=CTOK,
                                       opt=Opt(CoAPOptionContentFormat(0)),
                                       pl=Not(b"")))

        self.next_skip_ack(optional=True)

        self.chain()

        self.match_coap("client", CoAP(type="con", code="get",
                                       opt=self.uri("/multi-format", CoAPOptionAccept(41))))
        CMID2 = self.frame.coap["mid"]
        CTOK2 = self.frame.coap["tok"]
        if CMID2 is Not(b''):
            if CMID2 == CMID:
                self.setverdict("fail", "Message ID should be different")
        if CTOK2 is Not(b''):
            if CTOK2 == CTOK:
                self.setverdict("fail", "Token should be different")

        self.next_skip_ack()

        self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                       code=2.05,
                                       mid=CMID2,
                                       tok=CTOK2,
                                       opt=Opt(CoAPOptionContentFormat(41)),
                                       pl=Not(b"")))

        self.next_skip_ack(optional=True)



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from	ttproto.core.typecheck	import *
from	ttproto.core.data	import *

__all__ = [
	"SubtypeClass",
]

def SubtypeClass (pattern: is_data):

	pattern = as_data (pattern)
	base	= pattern.get_type()

	# FIXME: there is no way to automatically check that the value is valid
	#	 (e.g. at initialisation or when the object is updated)
	#	 -> would require some notification function when something is updated
	#	    (especially in the constructors)
	class SubtypeValue (base):
		@staticmethod
		def get_base_type():
			return base

		@classmethod
		def is_valid_value (cls, value):
			return super (SubtypeValue, cls).is_valid_value (value) and pattern.match (value)

	def SubtypeMetaclass (name, bases, classdict):
			assert len (bases) == 0 # FIXME: it could be useful to have multiple inheritance in some cases ?
			return type (name, (SubtypeValue,), classdict)

	return SubtypeMetaclass




import unittest
import logging
import json
from os import path

from ttproto.core.typecheck3000 import InputParameterError
from tests.test_tools.struct_validator import StructureValidator

from ttproto.core.packet import PacketValue
from ttproto.core.lib.inet.meta import InetPacketValue
from ttproto.core.lib.inet.coap import CoAP
from ttproto.core.lib.inet.sixlowpan import SixLowpan
from ttproto.core.dissector import Frame, Capture, ReaderError, get_dissectable_protocols
from ttproto.utils.pcap_filter import openwsn_profile_filter


class CaptureAndDissectionsTestCase(unittest.TestCase):
    """
    Test dissections. Unit testing methods
    """

    # #################### Tests parameters #########################

    # File path
    TEST_FILE_DIR = 'tests/test_dumps'

    # dissect CoAP pcap with other UDP messages:
    PCAP_FILE = path.join(TEST_FILE_DIR, 'coap', 'CoAP_plus_random_UDP_messages.pcap')

    # pcaps that MUST throw exceptions
    WRONG_TEST_FILE_DIR_NAME = 'others'
    EMPTY_PCAP_FILE = path.join(TEST_FILE_DIR, WRONG_TEST_FILE_DIR_NAME, 'empty_pcap.pcap')
    NOT_A_PCAP_FILE = path.join(TEST_FILE_DIR, WRONG_TEST_FILE_DIR_NAME, 'not_a_pcap_file.dia')

    # Create a struct checker object
    struct_validator = StructureValidator()

    # #################### Init and deinit functions #########################
    def setUp(self):
        """
            Initialize the dissector instance
        """
        self.capture = Capture(self.PCAP_FILE)

    # #################### Utilities functions #########################

    def check_summary(self, summary):
        self.assertTrue(type(summary), tuple)
        self.assertEqual(len(summary), 2)
        self.assertTrue(type(summary[0]), int)
        self.assertGreater(summary[0], 0)
        self.assertTrue(type(summary[1]), str)
        self.assertGreater(len(summary[1]), 0)

    # #################### Tests functions #########################

    # ##### get_dissectable_protocols
    def test_get_dissectable_protocols(self):

        # Get implemented protocols and check their values
        implemented_protocols = get_dissectable_protocols()
        self.assertEqual(type(implemented_protocols), list)
        self.assertGreater(len(implemented_protocols), 0)
        for prot in implemented_protocols:
            self.assertTrue(issubclass(prot, PacketValue))

    # ##### summary
    def test_summary_without_filtering(self):

        # Get and check the summary
        summary = self.capture.summary()
        self.assertTrue(type(summary), list)
        self.assertTrue(len(summary), 5)

        i = 1
        for f_sum in summary:
            self.check_summary(f_sum)
            self.assertEqual(f_sum[0], i)
            i += 1

        # Try to get another summary with None provided
        summary_with_none = self.capture.summary(None)
        self.assertEqual(summary, summary_with_none)

    def test_summary_with_filtering_on_coap(self):

        # Get and check the summary
        summary = self.capture.summary(CoAP)
        self.assertTrue(type(summary), list)
        self.assertTrue(len(summary), 2)

        i = 4  # CoAP frames are n°4 and 5
        for f_sum in summary:
            self.check_summary(f_sum)
            self.assertEqual(f_sum[0], i)
            i += 1

    def test_summary_with_filtering_on_protocols(self):

        # For every implemented protocols
        for prots in get_dissectable_protocols():

            # Get and check the summary
            summary = self.capture.summary(prots)
            self.assertTrue(type(summary), list)
            for f_sum in summary:
                self.check_summary(f_sum)

    def test_summary_with_filtering_on_none_type(self):

        # Get and check the summary
        with self.assertRaises(InputParameterError):
            summary = self.capture.summary(type(None))

    def test_summary_with_filtering_on_not_a_protocol(self):

        # Get and check the summary
        with self.assertRaises(InputParameterError):
            summary = self.capture.summary(Frame)

    def test_summary_with_wrong_pcap_file(self):

        # Create two wrong dissect instances, assert an exeption is raised
        with self.assertRaises(ReaderError):
            dis_wrong_file = Capture(self.NOT_A_PCAP_FILE)
            dis = dis_wrong_file.summary()
        with self.assertRaises(ReaderError):
            dis_empty_file = Capture(self.EMPTY_PCAP_FILE)
            dis = dis_empty_file.summary()

    # ##### dissect
    def test_dissection_without_filtering(self):

        # Get and check the dissect
        dissect = self.capture.get_dissection()
        self.assertTrue(type(dissect), list)
        self.assertTrue(len(dissect), 5)

        i = 1
        for frame in dissect:
            self.struct_validator.check_frame(frame)
            self.assertEqual(frame['id'], i)
            i += 1

        # Try to get another dissect with None provided
        dissect_with_none = self.capture.get_dissection(None)
        self.assertEqual(dissect, dissect_with_none)

    def test_dissection_with_filtering_on_coap(self):

        # Get and check the dissect
        dissect = self.capture.get_dissection(CoAP)
        self.assertTrue(type(dissect), list)
        self.assertTrue(len(dissect), 2)

        i = 4  # CoAP frames are n°4 and 5
        for frame in dissect:
            self.struct_validator.check_frame(frame)
            self.assertEqual(frame['id'], i)
            i += 1

    def test_dissection_with_filtering_on_protocols(self):

        # For every implemented protocols
        for prots in get_dissectable_protocols():

            # Get and check the dissect
            dissect = self.capture.get_dissection(prots)
            self.assertTrue(type(dissect), list)
            for frame in dissect:
                self.struct_validator.check_frame(frame)

    def test_dissection_with_filtering_on_none_type(self):

        # Get and check the dissect
        with self.assertRaises(InputParameterError):
            dissect = self.capture.get_dissection(type(None))

    def test_dissection_with_filtering_on_not_a_protocol(self):

        # Get and check the dissect
        with self.assertRaises(InputParameterError):
            dissect = self.capture.get_dissection(Frame)

    def test_dissection_with_wrong_pcap_file(self):

        with self.assertRaises(ReaderError):
            dis_wrong_file = Capture(self.NOT_A_PCAP_FILE)
            dis = dis_wrong_file.get_dissection()
        with self.assertRaises(ReaderError):
            dis_empty_file = Capture(self.EMPTY_PCAP_FILE)
            dis = dis_empty_file.get_dissection()


# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from ttproto.core.data import *
from ttproto.utils import pure_pcapy
import logging

from ttproto.core.typecheck import *
from ttproto.core.lib.ethernet import Ethernet
from ttproto.core.lib.encap import *
from ttproto.core.lib.inet.ipv6 import IPv6
from ttproto.core.lib.ieee802154 import Ieee802154
from ttproto.core.lib.readers.capture_reader import CaptureReader

log = logging.getLogger(__name__)
log.setLevel(level=logging.WARNING)

_map_link_type = {
    pure_pcapy.DLT_LINUX_SLL: LinuxCookedCapture,
    pure_pcapy.DLT_NULL: NullLoopback,
    pure_pcapy.DLT_EN10MB: Ethernet,
    pure_pcapy.DLT_IEEE802_15_4: Ieee802154,
    pure_pcapy.DLT_IEEE802_15_4_NOFCS: Ieee802154,
    # TODO DLT_RAW could be ipv4, need to do some introspection of packet before mapping this DLT
    pure_pcapy.DLT_RAW: IPv6
    # pure_pcapy.DLT_IEEE802_15_4_NONASK_PHY: Ieee802154
}


class PcapReader(CaptureReader):
    """
    Reader class for pcap capture files
    """

    @typecheck
    def __init__(self, file: str):
        """
        Initialize the reader with the corresponding file

        :param file: The path to the file to read
        :type file: str
        """

        self.__pcap_file = open(file, 'rb')
        try:
            self.__reader = pure_pcapy.Reader(self.__pcap_file)
        except Exception as e:
            self.__pcap_file.close()
            raise e
        log.debug("datalink: %d" % self.__reader.datalink())

        try:
            decode_type = _map_link_type[self.__reader.datalink()]
        except KeyError:
            decode_type = bytes

        self.__decode_type = get_type(decode_type)

    def __del__(self):
        # Close the file only if it was opened before
        try:
            self.__pcap_file.close()
        except:
            pass

    def next(self):

        h, b = self.__reader.next()

        if not h:
            return None

        # timestamp
        ts = h.getts()
        ts = ts[0] + ts[1] * 0.000001

        # decode the packet
        try:
            log.debug('Decoding bytes as %s: %s ' % (repr(b), self.__decode_type))
            m = Message(b, self.__decode_type)
            #log.debug(m.display())
            exc = None
        except Exception as e:
            m = Message(b)
            exc = e

        return ts, m, exc

    def __iter__(self):
        while True:
            f = self.next()

            if not f:
                return

            yield f



#!/usr/bin/env python3

from ..common import *


class TD_COAP_CORE_22(CoAPTestCase):
    """
---
TD_COAP_CORE_22:
    cfg: CoAP_CFG_BASIC
    obj: Perform GET transaction with responses containing the ETag option
        and requests containing the If-Match option (CON mode)
    pre:
    - Server offers a /validate resource
    - Client & server supports ETag and If-Match option
    - "The Client \u2018s cache must be purged"
    ref: '[COAP] 5.8.1, 5.10.7, 5.10.9, 12.1.12'
    seq:
    -   n: client gets the resource
    -   s: "Client is requested to send a confirmable GET request to server\
            \u2019s resource"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 1 (GET)
            - "Client-generated Message ID (\u2794 CMID)"
            - "Client-generated Token (\u2794 CTOK)"
            - Uri-Path option "validate"
            - No ETag option
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.05 (Content)
            - Message ID = CMID, Token = CTOK
            - "Option type = ETag, value = a value chosen by the server\
                \ (\u2794 ETAG1)"
            - Non-empty Payload
    -   n: single update
    -   s: "Client is requested to send a confirmable PUT request to server\
            \u2019s resource so as to perform an atomic update"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 3 (PUT)
            - "Another client-generated Message ID \u2260 CMID (\u2794 CMID2)"
            - "Client-generated Token which may or may not be \u2260 CTOK\
                \ (\u2794 CTOK2)"
            - Content-format option
            - Uri-Path option "validate"
            - Option type = If-Match, value = ETAG1 (ETag value received
                in step 3)
            - An arbitrary payload (which differs from the payload received
                in step 3)
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.04 (Changed)
            - Message ID = CMID2, Token = CTOK2
            - Content-format option if payload non-empty
            - Empty or non-empty Payload
    -   v: 'Client displays the response and the server changed its resource '
    -   n: concurrent updates
    -   s: "Client is requested to send a confirmable GET request to server\
            \u2019s resource"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 1 (GET)
            - "Another client-generated Message ID \u2260 CMID and \u2260\
                \ CMID2 (\u2794 CMID3)"
            - "Client-generated Token which may or may not be \u2260 CTOK\
                \ or CTOK2 (\u2794 CTOK3)"
            - Uri-Path option "validate"
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.05 (Content)
            - Message ID = CMID3, Token = CTOK3
            - "Option type = ETag, value = a value \u2260 ETAG1 chosen by\
                \ the server (\u2794 ETAG2)"
            - The Payload sent in step 5
    -   v: 'Client displays the response '
    -   s: "Update the content of the server\u2019s resource from a CoAP\
            \ client once more"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 3 (PUT)
            - "Another client-generated Message ID \u2260 CMID, CMID2, CMID3\
                \ (\u2794 CMID4)"
            - "Client-generated Token which may or may not be \u2260 CTOK,\
                \ CTOK2, CTOK3 (\u2794 CTOK4)"
            - Content-format option
            - Uri-Path option "validate"
            - An arbitrary payload (which differs from the payloads received
                in steps 3 and 10)
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.04 (Changed)
            - Message ID = CMID4, Token = CTOK4
            - Content-format option if payload non-empty
            - Empty or non-empty Payload
    -   v: 'Client displays the response and the server changed its resource '
    -   s: "Client is requested to send a confirmable PUT request to server\
            \u2019s resource so as to perform an atomic update, assuming it is\
            \ still unchanged from step 10"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 3 (PUT)
            - "Another client-generated Message ID \u2260 CMID, CMID2, CMID3\
                \ (\u2794 CMID4)"
            - "Client-generated Token which may or may not be \u2260 CTOK,\
                \ CTOK2, CTOK3 (\u2794 CTOK4)"
            - Content-format option
            - Uri-Path option "validate"
            - Option type = If-Match, value = ETAG2 (ETag value received
                in step 10)
            - An arbitrary payload (which differs from the previous payloads)
    -   c:
        - 'Server sends response containing:'
        -   - Code = 4.12 (Precondition Failed)
            - Message ID = CMID4, Token = CTOK4
            - Optional Content-format option
            - Empty or non-empty Payload
    -   v: Client displays the response and the server did not update the
            content of the resource
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. note::
            Check the number/value of the uri query options or not?
        """
        return [
            CoAP(type='con', code='get', opt=Opt(CoAPOptionUriPath("validate"))),  # Step 1
            CoAP(type='con', code='put', opt=Opt(CoAPOptionUriPath("validate"))),  # Step 4
            CoAP(type='con', code='get', opt=Opt(CoAPOptionUriPath("validate"))),  # Step 8
            CoAP(type='con', code='put', opt=Opt(CoAPOptionUriPath("validate"))),  # Step 12
            CoAP(type='con', code='put', opt=Opt(CoAPOptionUriPath("validate")))  # Step 13
        ]

    def run(self):
        # Preamble
        # Step 2
        self.match("client", CoAP(type="con",
                                  code="get",
                                  opt=All(
                                      Opt(CoAPOptionUriPath("validate")),
                                      NoOpt(CoAPOptionETag()),
                                  )))
        CMID = self.coap["mid"]
        CTOK = self.coap["tok"]

        self.next_skip_ack()

        # Step 3
        if not self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                         code=2.05,
                                         mid=CMID,
                                         tok=CTOK,
                                         opt=Opt(CoAPOptionETag()),
                                         pl=Not(b""))):
            raise self.Stop()

        ETAG1 = self.coap["opt"][CoAPOptionETag]["val"]
        pl_3 = self.coap["pl"]

        self.next_skip_ack(optional=True)

        # Part A

        # Step 5

        self.match("client", CoAP(type="con",
                                  code="put",
                                  opt=Opt(
                                      CoAPOptionContentFormat(),
                                      CoAPOptionUriPath("validate"),
                                      CoAPOptionIfMatch(ETAG1),
                                  ),
                                  pl=All(Not(b""), Not(pl_3))))
        CMID2 = self.coap["mid"]
        CTOK2 = self.coap["tok"]

        if CMID2 != b"":
            if CMID2 == CMID:
                self.set_verdict("fail", "Message ID should be different")
        if CTOK2 != b"":
            if CTOK2 == CTOK:
                self.set_verdict("fail", "Token should be different")
        pl_5 = self.coap["pl"]
        self.next_skip_ack()

        # Step 6
        if not self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                         code=2.04,
                                         mid=CMID2,
                                         tok=CTOK2)):
            raise self.Stop()
        if self.match("server", CoAP(pl=Not(b"")), None):
            self.match("server", CoAP(
                opt=Opt(CoAPOptionContentFormat()),
            ), "fail")

        self.next_skip_ack(optional=True)

        # Part B

        # Step 9

        self.match("client", CoAP(type="con",
                                  code="get",
                                  opt=Opt(CoAPOptionUriPath("validate"))))
        CMID3 = self.coap["mid"]
        CTOK3 = self.coap["tok"]

        if CMID3 != b"":
            if CMID3 == CMID or CMID3 == CMID2:
                self.set_verdict("fail", "Message ID should be different")
        if CTOK3 != b"":
            if CTOK3 == CTOK or CTOK3 == CTOK2:
                self.set_verdict("fail", "Token should be different")

        self.next_skip_ack()

        # Step 10
        if not self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                         code=2.05,
                                         mid=CMID3,
                                         tok=CTOK3,
                                         opt=Opt(CoAPOptionETag(Not(ETAG1))),
                                         pl=pl_5)):
            raise self.Stop()

        ETAG2 = self.coap["opt"][CoAPOptionETag]["val"]
        pl_10 = self.coap["pl"]

        self.next_skip_ack(optional=True)

        # Step 13
        self.match("client", CoAP(type="con",
                                  code="put",
                                  opt=Opt(
                                      CoAPOptionUriPath("validate"),
                                      CoAPOptionContentFormat(),
                                  ),
                                  pl=All(Not(b""), Not(pl_3), Not(pl_10))))
        CMID_step13 = self.coap["mid"]
        CTOK_step13 = self.coap["tok"]

        if CMID_step13 != b"":
            if CMID_step13 == CMID or CMID_step13 == CMID2 or CMID_step13 == CMID3:
                self.set_verdict("fail", "Message ID should be different")
        if CTOK_step13 != b"":
            if CTOK_step13 == CTOK or CTOK_step13 == CTOK2 or CTOK_step13 == CTOK3:
                self.set_verdict("fail", "Token should be different")
        pl_13 = self.coap["pl"]

        self.next_skip_ack()

        # Step 13 bis
        self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                  code=2.04,
                                  mid=CMID_step13,
                                  tok=CTOK_step13, ))

        if self.match("server", CoAP(pl=Not(b"")), None):
            self.match("server", CoAP(opt=Opt(CoAPOptionContentFormat()), ), "fail")

        self.next_skip_ack(optional=True)

        # Step 14

        self.match("client", CoAP(type="con",
                                  code="put",
                                  opt=Opt(
                                      CoAPOptionContentFormat(),
                                      CoAPOptionUriPath("validate"),
                                      CoAPOptionIfMatch(ETAG2),
                                  ),
                                  pl=All(Not(b""), Not(pl_13))))
        CMID4 = self.coap["mid"]
        CTOK4 = self.coap["tok"]

        if CMID4 != b"":
            if CMID4 == CMID or CMID4 == CMID2 or CMID4 == CMID3 or CMID4 == CMID_step13:
                self.set_verdict("fail", "Message ID should be different")
        if CTOK4 != b"":
            if CTOK4 == CTOK or CTOK4 == CTOK2 or CTOK4 == CTOK3 or CTOK4 == CTOK_step13:
                self.set_verdict("fail", "Token should be different")

        self.next_skip_ack()
        # Step 15
        self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                  code=4.12,
                                  mid=CMID4,
                                  tok=CTOK4), 'fail')



import unittest
import os
import json
import logging

from ttproto.core.lib.inet.coap import CoAP

from tests.test_tools.struct_validator import StructureValidator
from tests.test_dissector.test_dissect_pcaps import DissectPcapTestCase


class DissectorTestCase_coap(DissectPcapTestCase):
    """
    Test class for the dissector tool

    python3 -m unittest tests.test_dissector.test_dissector_coap.DissectorTestCase.test_that_it_can_dissect_all_pcaps -vvv
    """

    # #################### Tests parameters #########################

    # File path
    PCAP_FILES_DISSECTION_DIRS = [
        'tests/test_dumps/coap',
        'tests/test_dumps/coap_core',
        'tests/test_dumps/coap_block',
        'tests/test_dumps/coap_observe',
    ]
    PROTO_CLASS_FILTER = CoAP
    TMP_DIR = 'tmp/'

    # Create a struct checker object
    struct_validator = StructureValidator()


# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()



from ..common import *


class TD_COAP_OBS_09(CoAPTestCase):
    """
---
TD_COAP_OBS_09:
    obj: Update of the observed resource
    cfg: CoAP_CFG_BASIC
    ref: '[OBSERVE] 4.2.3'

    pre:
        - Client supports Observe option
        - Server supports Observe option
        - "Server offers an observable resource /obs which changes periodically
        (e.g. every 5s) which produces confirmable notifications"

    seq:
    -   s: 'Client is requested to send to the server a confirmable GET
        request with observe option for resource /obs'

    -   c:
        - 'The request sent by client contains:'
        -   - Type = 0 (CON)
            - Code = 1 (GET)
            - Token value = a value generated by the client
            - Observe option = empty

    -   c:
        - 'Server sends the response containing:'
        -   - Type = 2 (ACK)
            - Code = 2.05 (Content)
            - Content-format of the resource /obs
            - Token value = same as one found in the step 2
            - Observe option with a sequence number

    -   c:
        - 'Server sends a notification containing:'
        -   - Type = 0 (CON)
            - Code = 2.05 (Content)
            - Content-format = same as one found in the step 3
            - Token value = same as one found in the step 2
            - Observe option indicating increasing values

    -   c: Client displays the received information

    -   c: Client sends an ACK

    -   s: 'Update the /obs resource of the server\u2019s resource with a new payload
            having the same Content-Format (either locally or by having another
            CoAP client perform a PUT request)'

    -   c:
        - 'Server notifications contains:'
        -   - Type = 0 (CON)
            - Code = 2.05 (Content)
            - Content-format = same as one found in the step 3
            - Token value = same as one found in the step 2
            - Observe option indicating increasing values
            - Payload = the new value sent at step 7

    -   v: Client displays the new value of /obs sent in step 8

    -   c: Client sends an ACK

    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [
            CoAP(type='con', code='get', opt=Opt(CoAPOptionUriPath("obs"), CoAPOptionObserve(0))),
            CoAP(type='con', code='post', opt=Opt(CoAPOptionUriPath("obs"), CoAPOptionContentFormat()))
        ]

    def run(self):
        request = CoAP(type="con", code="get", opt=self.uri("/obs", CoAPOptionObserve (0)))
        response = CoAP(type="ack", code=2.05, opt=Opt(CoAPOptionObserve(), CoAPOptionContentFormat()))

        # Step 2
        self.match("client", request)

        token = self.coap["tok"]
        uri = self.coap.get_uri()

        self.next()

        # Step 3
        if not self.match("server", response):
            raise self.Stop()
        self.match("server", CoAP(tok = token), "fail")

        content_format = self.coap["opt"][CoAPOptionContentFormat]["val"]
        index          = self.coap["opt"][CoAPOptionObserve]["val"]
        self.next()
        pl = None

        rst_sent_by_client = False

        # we need to observe at least one notification
        # (to ensure that the server is observing the resource for the client)
        notifications_received_from_server = 0

        ack_sent_by_client = 0

        # Step 4/6 and step 8/10
        while self.match("server", CoAP(type="con", code=2.05,
                        opt=Opt(CoAPOptionObserve())), None):
            self.match("server", CoAP(tok = token), "fail")
            self.match("server", CoAP(opt = Opt(CoAPOptionContentFormat(content_format))), "fail")
            new_index = self.coap["opt"][CoAPOptionObserve]["val"]
            pl = self.coap["pl"]

            self.set_verdict(("pass" if new_index > index else "fail"),
                "value of observe option must be increasing")
            self.next()
            notifications_received_from_server += 1

            # Step 6
            if self.match("client", CoAP(type="rst", code=0), None):
                rst_sent_by_client = True
                self.set_verdict("pass", "End of communication from the client.")
                self.next(True)

            # We do a while loop here if there is several following ACK in the capture.
            # It may happen sometime, e.g because of latency.
            if self.match("client", CoAP(type="ack", code=0), None):
                ack_sent_by_client += 1
                self.set_verdict("pass", "ACK from client.")
                self.next(True)

            #self.next()
            stimulis = self.__check_stimulis(content_format)

        # TODO Check which code to check to throw the error.
        if self.match("server", CoAP(type="con", tok = token, code=Any(134,160,140,143)), None):
            self.set_verdict("fail", "Server sent an error (decimal code = %d) to the client." % self.coap['code'])
        else:
            self.set_verdict("pass", "Server did not send error to the client after payload changed.")
        if notifications_received_from_server < 1:
            self.set_verdict("inconclusive", "Only one notification since the beginning of the test.\
            Unable to perform the check at step 8.")

        delta = notifications_received_from_server - ack_sent_by_client
        if delta == 0:
            self.set_verdict("pass", "Every notifications has been acknoledged.")
        elif delta == 1 and rst_sent_by_client:
            self.set_verdict("pass", "Every notifications has been acknoledged. \
excepted the last one as the client terminated the test session properly with an RST.")
        raise self.Stop()

    def __check_stimulis(self, content_format:str)->bool:
        # Step 7: Stimulus. This step is optional as changing the Content-Format
        # may be done internally by the server.
        # return a boolean that determine if we encountered this optional stimulis
        stimulis = False
        if self.match("client", CoAP(type='con', code='post', opt=Opt(CoAPOptionUriPath("obs"), CoAPOptionContentFormat())), None):
            new_content_format = self.coap["opt"][CoAPOptionContentFormat]["val"]
            self.set_verdict(("pass" if new_content_format == content_format
                else "inconclusive"),"Stimulus: POST performed with the same Content-Format.\
                If the POST is performed with another Content-Format, we have\
                no way to check the behavior of this TC")
            stimulis = True
            self.next()

        if self.match("server", CoAP(type="ack", code=2.04), None):
            self.set_verdict("pass", "Stimulus: Confirmation from server that POST suceeded with the same Content-Format")
            self.next()
            stimulis = True

        return stimulis



from ..common import *


class TD_LWM2M_1_INT_204(CoAPTestCase):
    """
---
testcase_id: TD_LWM2M_1.0_INT_204
uri : http://openmobilealliance.org/iot/lightweight-m2m-lwm2m
configuration: LWM2M_CFG_01
objective: 
  - Quering the Resources values of Device Object (ID:3) on the Client in JSON format
  - - Manufacturer Name (id:0)
    - Model number (ID:1)
    - Serial number (ID:2)
    - Firware Version (ID:3)
    - Error Code (ID:11)
    - Supported Binding and Modes (ID:16)
pre_conditions:
  - Device is registred at the LWM2M server
  - The Client supports the minimum Configuration C.1.
references: 'OMA-ETS-LightweightM2M-V1_0_1-20170926-A'
sequence:
  - step_id: 'TD_LWM2M_1.0_INT_204_step_01'
    type: stimuli
    node : lwm2m_server
    description:
      - 'Server sends a READ request (COAP GET) on device object instance'
      - - Type = 0 (CON)
        - Code = 1 (GET)

  - step_id: 'TD_LWM2M_1.0_INT_204_step_02'
    type: check
    description:
      - 'The request sent by the server contains'
      - - Type=0 and Code=1
        - Accept option = application/vnd.oma.lwm2m+json
        - Uri-Path option = 3/0


  - step_id: 'TD_LWM2M_1.0_INT_204_step_03'
    type: check
    description:
        - 'Client sends response containing'
        - - Code = 2.05 (Content)
          - content-format option=application/vnd.oma.lwm2m+json
          - Non-empty payload= contains the expected values of Manufacturer Name, Model number, Serial number, Firmware Version, error code, and supported Binding and modes resources. 

  - step_id: 'TD_LWM2M_1.0_INT_204_step_04'
    type: verify
    node: lwm2m_server
    description:
        - 'Requested data is successfully displayed'

    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [CoAP(type='con', code='get')]

    def run(self):
        self.match('server', CoAP(type='con', code='get', opt=self.uri('/3/0')), 'fail')
        self.match('server', CoAP(pl=(b'')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionAccept('11543'))), 'fail')
        
        self.next()

        self.match('client', CoAP(code=2.05, pl=Not(b'')), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat('11543'))), 'fail')

        validation = validate(str(self.coap['pl']),'3')
        self.set_verdict(validation[0], validation[1])




from ttproto.tat_coap.testcases.td_coap_core_19 import TD_COAP_CORE_19


class TD_COAP_CORE_25 (TD_COAP_CORE_19):
    """
---
TD_COAP_CORE_25:
    cfg: CoAP_CFG_03
    obj: Perform POST transaction with responses containing several
         Location-Query option (Reverse proxy)
    pre:
        - Proxy is configured as a reverse-proxy for the server
        - Proxy’s cache is cleared
        - Server accepts creation of new resource on uri  /location-query, the
          location of the created resource contains two query parameters
          ?first=1&second=2
    ref: '[1] clause 5.8.1,5.10.8,5.9.1.1, 8.2.2,8.2.1,10.2.2,11.2'
    seq:
        -   s: Client is requested to send a confirmable POST request to proxy
        -   c: Proxy receives the request from client  & forwards it to
               server's resource
        -   c:
            - 'Forwarded request must contain:'
            -   - Type = 0 (CON)
                - Code = 2 (POST)
                - An arbitrary payload
                - Content-format option
        -   c:
            - 'Server sends response to proxy containing:'
            -   - Code = 65 (2.01 created)
                - Two options whose type is Location-Query
                - The first option contains first=1
                - The second option contains second=2
        -   c: Proxy forwards the response to client
        -   c: Client displays the message
        -   v:
            - 'Client interface returns the response:'
            -   - 2.01 created
                - Location: coap://proxy/?first=1&second=2
    """

    reverse_proxy = True



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import atexit, threading, tempfile, os, signal, socket, time, weakref, traceback

from ttproto.core.typecheck	import *
from ttproto.core.data		import *
from ttproto.core		import exceptions, port

class SocatMessagePort (port.RawMessagePort):

	@typecheck
	def __init__ (self, socat_endpoint: str, decode_type: is_type = bytes, endpoint: optional (port.BaseMessagePort) = None):
		port.RawMessagePort.__init__ (self, decode_type, endpoint)

		tmpdir = tempfile.mkdtemp()
		self.__sockname = os.path.join (tmpdir, "sock_my")
		self.__sockname_socat = os.path.join (tmpdir, "sock_socat")
		self.__kill_event = threading.Event()

		# create the socket
		self.__sock = socket.socket (socket.AF_UNIX, socket.SOCK_DGRAM)
		self.__sock.bind (self.__sockname)

		# launch socat
		self.__pid = os.fork()
		if not self.__pid:
			# child
			os.execl ("/usr/bin/socat", "socat",
					socat_endpoint,
					"UNIX-CONNECT:%s,bind=%s,type=%d" % (self.__sockname, self.__sockname_socat, socket.SOCK_DGRAM))
		retry = 10
		while True:
			try:
				time.sleep(0.1)
				self.__sock.connect (self.__sockname_socat)
				break
			except socket.error as e:
				retry -= 1
				if retry == 0:
					raise e

		# launch a listening thread
		self.__thread = threading.Thread (target = self.__thread_func)
		# this is a daemon thread (will not block at exit)
		self.__thread.daemon = True
		self.__thread.start()

		# register an exit handler to free the resources when leaving
		atexit.register (self.kill)

	@typecheck
	def enqueue (self, msg: Message):
		assert isinstance (msg.get_binary(), bytes) # supports only byte-aligned data
		try:
			self.__sock.send (msg.get_binary())
		except Exception as e:
			if not self.__kill_event.is_set():
				print("Cannot send the message")
				traceback.print_exc()
				#FIXME: should have an error handler that reports an error verdict

	def __thread_func (self):
		try:
			while True:
				buff = self.__sock.recv (65535)
				if len (buff) == 0 and self.__kill_event.is_set():
					# the port is being killed -> just return silently
					return

				# forward it
				self._forward (buff)
				#TODO: warning if there is no endpoint ?

		except Exception as e:
			print("Error in receiving thread")
			traceback.print_exc()

			#TODO: report the error to the tester

	def kill (self):
		# set the kill event (to notify other threads)
		self.__kill_event.set()

		# close the socket (this will terminate the listening thread)
		self.__sock.shutdown(socket.SHUT_RDWR)

		# stop the socat process
		os.kill (self.__pid, signal.SIGTERM)
		os.waitpid (self.__pid, 0)

		# remove the tmp files
		os.unlink (self.__sockname)
		os.unlink (self.__sockname_socat)
		os.rmdir (os.path.dirname (self.__sockname))

		# unregister our exit handler
		atexit.unregister (self.kill)




from ..common import *


class TD_COAP_CORE_10(CoAPTestCase):
    """
---
TD_COAP_CORE_10:
    cfg: CoAP_CFG_BASIC
    obj: Perform GET transaction containing non-empty Token (CON mode)
    pre: Server offers a /test resource with resource content is not empty
        that handles GET
    ref: '[COAP] 2.2, 5.8.1, 5.10.1'
    seq:
    -   s: "Client is requested to send a GET request to server\u2019s resource
            \ with non-empty Token option"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 1 (GET)
            - "Client-generated Message ID (\u2794 CMID)"
            - "Client-generated Token (\u2794 CTOK)"
            - Length of the token should be between 1 to 8 Bytes
            - Uri-Path option "test"
    - &id003
        c:
        - 'Server sends response containing:'
        -   - Code = 2.05 (Content)
            - Message ID = CMID, Token = CTOK
            - Content-format option
            - Non-empty Payload
    -   v: Client displays the response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        # TODO client message w/ token length = 0 -> fail?
        return [
            CoAP(type='con', code='get', tok=Not(b''),
                 opt=Opt(CoAPOptionUriPath("test")))
        ]

    def run(self):
        self.match("client", CoAP(code="get", type="con", tok=Not(b""), opt=self.uri("/test")))
        self.match("client", CoAP(tok=Length(bytes, (1, 8))), "fail")
        CMID = self.coap["mid"]
        CTOK = self.coap["tok"]

        self.next()

        if self.match("server", CoAP(code=2.05, pl=Not(b""))):
            self.match("server", CoAP(mid=CMID, tok=CTOK, opt=Opt(CoAPOptionContentFormat())), "fail")



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from    ttproto.core.data import *
from    ttproto.core.typecheck import *
from    ttproto.core.lib.inet.meta import *
from    ttproto.core.lib.inet.basics import *
from    ttproto.core.lib.inet.ipv6 import *
from    ttproto.core import exceptions

import ttproto.core.lib.inet.ipv6

__all__ = [
    'ICMPv6',
    'ICMPv6Option',
    'ICMPv6OptionList',
    'ICMPv6LLOption',
    'ICMPv6SLLOption',
    'ICMPv6TLLOption',
    'ICMPv6SLL',
    'ICMPv6TLL',
    'ICMPv6PIOption',
    'ICMPv6PI',
    'ICMPv6NeighborSolicitation',
    'ICMPv6NeighborAdvertisement',
    'ICMPv6NSol',
    'ICMPv6NAdv',
    'ICMPv6RouterSolicitation',
    'ICMPv6RouterAdvertisement',
    'ICMPv6RSol',
    'ICMPv6RAdv',
    'ICMPv6EchoRequest',
    'ICMPv6EchoReply',
    'ICMPv6EReq',
    'ICMPv6ERep',
    'ICMPv6DestinationUnreacheable',
    'ICMPv6Unre',
]


class _ICMPv6OptLengthDescription:
    def __getitem__(self, item):
        return "%d bytes" % (item * 8)


"""
defined in [RFC4443]

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |     Type      |     Code      |          Checksum             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      +                         Message Body                          +
      |                                                               |


"""


class ICMPv6(
    metaclass=InetPacketClass,
    fields=[
        ("Type", "type", UInt8, InetVariant()),
        ("Code", "code", UInt8, 0),
        ("Checksum", "chk", Hex(UInt16), InetIPv6Checksum()),
        ("Payload", "pl", Value)
    ],
    descriptions={
        "Type": {
            # source: http://www.iana.org/assignments/icmpv6-parameters
            1: "Destination Unreachable",
            2: "Packet Too Big",
            3: "Time Exceeded",
            4: "Parameter Problem",
            128: "Echo Request",
            129: "Echo Reply",
            130: "Multicast Listener Query",
            131: "Multicast Listener Report",
            132: "Multicast Listener Done",
            133: "Router Solicitation",
            134: "Router Advertisement",
            135: "Neighbor Solicitation",
            136: "Neighbor Advertisement",
            137: "Redirect Message",
            138: "Router Renumbering",
            139: "ICMP Node Information Query",
            140: "ICMP Node Information Response",
            141: "Inverse Neighbor Discovery Solicitation Message",
            142: "Inverse Neighbor Discovery Advertisement Message",
            143: "Version 2 Multicast Listener Report",
            144: "Home Agent Address Discovery Request Message",
            145: "Home Agent Address Discovery Reply Message",
            146: "Mobile Prefix Solicitation",
            147: "Mobile Prefix Advertisement",
            148: "Certification Path Solicitation Message",
            149: "Certification Path Advertisement Message",
            150: "Seamoby",
            151: "Multicast Router Advertisement",
            152: "Multicast Router Solicitation",
            153: "Multicast Router Termination",
            154: "FMIPv6 Messages",

            # defined in [I-D.6lowpan-nd-15]: (FIXME)
            31: "Address Registration Option (TBD1)",
            32: "6LoWPAN Context Option (TBD2)",
            33: "Authoritative Border Router Option (TBD3)",

        },
    }):
    @typecheck
    def find_type(self: is_flat_value, type_: is_type) -> optional(Value):
        result = super().find_type(type_)
        if result is not None:
            return result

        # try in the option list
        try:
            options = self["Options"]
        except exceptions.UnknownField:
            return None

        return None if options is None else options.find_type(type_)


"""
defined in [RFC4861]:

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |     Type      |    Length     |              ...              |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       ~                              ...                              ~
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""


class ICMPv6Option(
    metaclass=InetPacketClass,
    fields=[
        ("Type", "type", UInt8, InetVariant()),
        ("Length", "len", UInt8, InetLength(unit=8)),
        ("Value", "val", bytes, b""),
    ],
    descriptions={
        "Type": {
            # Registry Name: IPv6 Neighbor Discovery Option Formats
            # source: http://www.iana.org/assignments/icmpv6-parameters
            1: "Source Link-layer Address",
            2: "Target Link-layer Address",
            3: "Prefix Information",
            4: "Redirected Header",
            5: "MTU",
            6: "NBMA Shortcut Limit Option",
            7: "Advertisement Interval Option",
            8: "Home Agent Information Option",
            9: "Source Address List",
            10: "Target Address List",
            11: "CGA option",
            12: "RSA Signature option",
            13: "Timestamp option",
            14: "Nonce option",
            15: "Trust Anchor option",
            16: "Certificate option",
            17: "IP Address/Prefix Option",
            18: "New Router Prefix Information Option",
            19: "Link-layer Address Option",
            20: "Neighbor Advertisement Acknowledgment Option",
            23: "MAP Option",
            24: "Route Information Option",
            25: "Recursive DNS Server Option",
            26: "RA Flags Extension Option",
            27: "Handover Key Request Option",
            28: "Handover Key Reply Option",
            29: "Handover Assist Information Option",
            30: "Mobile Node Identifier Option",
            31: "DNS Search List Option",
            32: "Proxy Signature (PS)",
            138: "CARD Request option",
            139: "CARD Reply option",
            253: "RFC3692-style Experiment 1",
            254: "RFC3692-style Experiment 2",
        },
        "Length": _ICMPv6OptLengthDescription(),
    }):
    pass


class ICMPv6OptionList(
    metaclass=InetUnorderedListClass,
    content_type=ICMPv6Option):
    pass


"""
defined in [RFC4861]:

      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |     Type      |    Length     |    Link-Layer Address ...
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""


class ICMPv6LLOption(
    metaclass=InetPacketClass,
    variant_of=ICMPv6Option,
    prune=-1,
    fields	= [
		("LinkLayerAddress",	"hw",	ICMPv6LLAddress, b""),
	]):
	pass

class ICMPv6SLLOption (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6LLOption,
	id         = 1,
	):
	pass

class ICMPv6TLLOption (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6LLOption,
	id         = 2,
	):
	pass
"""
defined in [RFC4861]:
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |     Type      |    Length     | Prefix Length |L|A| Reserved1 |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Valid Lifetime                        |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                       Preferred Lifetime                      |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                           Reserved2                           |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      +                                                               +
      |                                                               |
      +                            Prefix                             +
      |                                                               |
      +                                                               +
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""

class ICMPv6PIOption (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6Option,
	prune      = -1,
	id         = 3,
	fields     = [
		("PrefixLength", 	"pflen" ,UInt8,		64),
		("L",			"l",	bool,		1),
		("A",			"a",	bool,		1),
		("Reserved1", 		"rsv1",	Bin (UInt6),	0),
		("ValidLifetime",	"vlt",	UInt32,		2592000),
		("PreferedLifetime",	"plt",	UInt32,		604800),
		("Reserved2", 		"rsv2",	Hex (UInt32),	0),
		("Prefix",		"pf",	IPv6Address,	"2001:DB8::"),
	]):
	pass

"""
defined in [RFC4443]:

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |     Type      |     Code      |          Checksum             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |           Identifier          |        Sequence Number        |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |     Data ...
      +-+-+-+-+-

"""
class ICMPv6Echo (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6,
	prune      = -1,
	fields     = [
		("Identifier",		"id",	Hex (UInt16),	0),
		("SequenceNumber",	"seq",	UInt16, 	0),
		("Payload", 		"pl",	Value),
	]):
	pass

class ICMPv6EchoRequest (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6Echo,
	id         = 128,		): pass


class ICMPv6EchoReply (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6Echo,
	id         = 129,		): pass

"""
defined in [RFC4443]:

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |     Type      |     Code      |          Checksum             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                             Unused                            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                    As much of invoking packet                 |
      +                as possible without the ICMPv6 packet          +
      |                exceeding the minimum IPv6 MTU [IPv6]          |

"""
class ICMPv6DestinationUnreacheable (
	metaclass	= InetPacketClass,
	variant_of	= ICMPv6,
	prune		= -1,
	id	   	= 1,
	fields		= [
		("Unused",		"un",	Hex (UInt32),	0),
		("Payload",		"pl",	IPv6,		0),
	],
	descriptions = {
	"Code": {
		0:	"No route to destination",
		1:	"Communication with destination administratively prohibited",
		2:	"Beyond scope of source address",
		3:	"Address unreachable",
		4:	"Port unreachable",
		5:	"Source address failed ingress/egress policy",
		6:	"Reject route to destination",
	}}):
	pass


"""
defined in [RFC4861]:
      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |     Type      |     Code      |          Checksum             |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |                           Reserved                            |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |                                                               |
     +                                                               +
     |                                                               |
     +                       Target Address                          +
     |                                                               |
     +                                                               +
     |                                                               |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |   Options ...
     +-+-+-+-+-+-+-+-+-+-+-+-

"""
class ICMPv6NeighborSolicitation (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6,
	prune      = -1,
	id         = 135,
	fields     = [
		("Reserved",		"rsv",	Hex (UInt32),	0),
		("TargetAddress",	"tgt",	IPv6Address,	"::"),
		("Options",		"opt",	ICMPv6OptionList, []),
	]):
	pass

"""
defined in [RFC4861]:

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |     Type      |     Code      |          Checksum             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |R|S|O|                     Reserved                            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      +                                                               +
      |                                                               |
      +                       Target Address                          +
      |                                                               |
      +                                                               +
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |   Options ...
      +-+-+-+-+-+-+-+-+-+-+-+-

"""
class ICMPv6NeighborAdvertisement (
	metaclass  = InetPacketClass,
	variant_of = ICMPv6,
	prune      = -1,
	id         = 136,
	fields     = [
		("Flags", 		"fl",	Hex (UInt16),	0),	# TODO: define the bits
		("Reserved", 		"rsv",	Hex (UInt16),	0),
		("TargetAddress",	"tgt",	IPv6Address,	"::"),
		("Options",		"opt",	ICMPv6OptionList, []),
	]):
	pass

"""
defined in [RFC4861]:

      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |     Type      |     Code      |          Checksum             |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |                            Reserved                           |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |   Options ...
     +-+-+-+-+-+-+-+-+-+-+-+-

"""
class ICMPv6RouterSolicitation (
	metaclass	= InetPacketClass,
	variant_of	= ICMPv6,
	prune		= -1,
	id 		= 133,
	fields		= [
		("Reserved",		"rsv",	Hex (UInt32),	0),
		("Options",		"opt",	ICMPv6OptionList, []),
	]):
	pass

"""
defined in [RFC4861]:

      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |     Type      |     Code      |          Checksum             |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     | Cur Hop Limit |M|O|  Reserved |       Router Lifetime         |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |                         Reachable Time                        |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |                          Retrans Timer                        |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |   Options ...
     +-+-+-+-+-+-+-+-+-+-+-+-

"""
class ICMPv6RouterAdvertisement (
	metaclass	= InetPacketClass,
	variant_of	= ICMPv6,
	prune		= -1,
	id 		= 134,
	fields		= [
		("CurHopLimit",		"hp",	UInt8,		0),
		("Managed",		"m",	bool,		0),
		("Other",		"o",	bool,		0),
		("HomeAgent",		"ha",	bool,		0),
		("RouterPreference",	"rp",	UInt2,		0),
		("Proxied",		"p",	bool,		0),
		("Reserved",		"rsv",	UInt2,		0),
		("RouterLifetime",	"rlt",	UInt16,		1800),
		("ReachableTime",	"tm",	UInt32,		0),
		("RetransTimer",	"rt",	UInt32,		0),
		("Options",		"opt",	ICMPv6OptionList, []),
	]):
	pass


# Aliases
ICMPv6NSol = ICMPv6NeighborSolicitation
ICMPv6NAdv = ICMPv6NeighborAdvertisement
ICMPv6RSol = ICMPv6RouterSolicitation
ICMPv6RAdv = ICMPv6RouterAdvertisement
ICMPv6EReq = ICMPv6EchoRequest
ICMPv6ERep = ICMPv6EchoReply
ICMPv6Unre = ICMPv6DestinationUnreacheable
ICMPv6SLL = ICMPv6SLLOption
ICMPv6TLL = ICMPv6TLLOption
ICMPv6PI = ICMPv6PIOption

# tell the ipv6 module that ip next header 58 should be mapped to the ICMPv6 class
ttproto.core.lib.inet.ipv6.ip_next_header_bidict.update({
	58:	ICMPv6,
})



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_18 (CoAPTestcase):
    """Identifier:
TD_COAP_CORE_18
Objective:
Perform POST transaction with responses containing several Location-Path options
(CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 5.8.1,5.10.8,5.9.1.1

Pre-test
conditions:
•	Server accepts creation of new resource on  /testand the created resource is
located at  /location1/location2/location3 (resource does not exist yet)

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a confirmable POST request to
server’s resource


2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 2 (POST)
•	An arbitrary payload
•	Content-format option
•	Uri-Path option "test"

3
Check
Server sends response containing:
•	Code = 65 (2.01 created)
and three options of type Location-Path, with the values (none of which contains a "/"):
•	location1
•	location2
•	location3



4
Verify
Client displays the response
"""
    request_uri = "/test"

    def run (self):
        self.match_coap ("client", CoAP (type="con", code = "post",pl=Not(b""),
                    opt=self.uri(
                        self.request_uri,
                        CoAPOptionContentFormat(),
                    )))

        self.next_skip_ack()

        self.match_coap ("server", CoAP (code=2.01,
                        opt=Opt(
                            CoAPOptionLocationPath ("location1"),
                            CoAPOptionLocationPath ("location2"),
                            CoAPOptionLocationPath ("location3"),
                )))

        self.next_skip_ack (optional = True)




#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
import numbers, threading, traceback
from .common import *

from ttproto.core import	clock, snapshot, exceptions

from ttproto.core.templates import Superset #FIXME: why we need it ?!

__all__ = [
	"SixLowpanImplementation",
]

class Scheduler (snapshot.EventSource):
	class Event:
		@typecheck
		def __init__ (self, process, name: optional(str)):
			self.process = process
			self.name = name


	def __init__ (self, clk: optional (clock.Clock) = None):
		super().__init__()

		self.__clock = clk if clk is not None else clock.Clock.get_instance()
		self.event = snapshot.Event (self, "event")

		self.__events = []
		self.__expired_events = []
		
	
	def reset (self):
		with self.lock:
			for e in self.__events:
				self.__clock.unschedule_events (e)
			self.__events  = []
			self.__expired_events = []
	
	@typecheck
	def __find_event (self, process, name: str):
		for e in self.__events:
			if e.process == process and e.name == name:
				return e
	
	@typecheck
	def __unschedule_event (self, process, name: str):
		ev = self.__find_event (process, name)
		if ev is not None:
			self.__clock.unschedule_events (ev)
			self.__events.remove (ev)
			try:
				self.__expired_events.remove (ev)
			except ValueError:
				pass
	
	@typecheck
	def __prepare_schedule_event (self, process, name: optional(str) = None):
		if name is not None:
			self.__unschedule_event (process, name)

		ev = self.Event (process, name)
		self.__events.append (ev)
		def callback():
			with self.lock:
				self.__expired_events.append (ev)
			
		return callback
	
	@typecheck
	def schedule_event_relative (self, time: numbers.Number, process, name: optional(str) = None):
		with self.lock:
			self.__clock.schedule_event_relative (time, self.__prepare_schedule_event (process, name))

	@typecheck
	def schedule_event_absolute (self, time: numbers.Number, process, name: optional(str) = None):
		with self.lock:
			self.__clock.schedule_event_absolute (time, self.__prepare_schedule_event (process, name))
	
	def evaluate_snapshot(self):
		self.__snapshot_expired_events = len (self.__expired_events)

	def time (self):
		return self.__clock.time()
	
	class __EventMatch (snapshot.EventMatch):
		def __init__ (self, snapshot_event, ev):
			super().__init__ (snapshot_event)
			self.__event = ev

		def get_ev (self):
			return self.__event

	def match_event (self):

		while self.__snapshot_expired_events:

			self.__snapshot_expired_events -= 1
			ev = self.__expired_events.pop(0)
			try:
				self.__events.remove (ev)
			except ValueError:
				#print ("WARNING: FIXME: unscheduled event still present")
				pass
			else:
				return self.__EventMatch (self.event, ev)
	
		return None


class SixLowpanImplementation (snapshot.SnapshotManager.Thread):

	class ConfigClass:
		pass
	config = ConfigClass()
	config.debug = 1
	config.eui64 = "0200000001020304"
	config.routeraddr = ""

	#delay we use when we schedule a RS
	config.DelayRS = 0.1

	#delay we use when we schedule a NS
	config.DelayNS = 0.1

	# time to refresh before a router expire
	config.routerRefreshTime = 6
	# time to refresh before an address expire
	config.AddressRefreshTime = 6

	class Process:
	
		def __init__ (self, implem, func, *k, **kw):
			"""Initialise the new process
			
			- implem	the implementation (python object) hosting this process 
			- function	generator function implementing the behaviour of this process

			"""
			self.__generator  = func (self, *k, **kw)
			self.__scheduler  = implem.get_scheduler()
			next (self.__generator)

		def send (self, msg):
			"""Report an event (message) to the process"""

			return self.__generator.send (msg)

		@typecheck
		def schedule_event_relative (self, time: numbers.Number, name: optional(str) = None):
			"""Schedule an event in the future.

			- time		the time at which the event is to be reported
			- name		name of the event (optional)

			The time is given in seconds starting from now.

			The name is optional. If there is already an existing
			event with the same name for the same process, then
			this existing event is overwritten by the new one.		
			"""
			self.__scheduler.schedule_event_relative (time, self, name)

		@typecheck
		def schedule_event_absolute (self, time: numbers.Number, name: optional(str) = None):
			"""Schedule an event in the future.

			- time		the time at which the event is to be reported
			- name		name of the event (optional)

			The time is given in seconds starting from the epoch (1st Jan 1970).
ig.delayRS, 
			The name is optional. If there is already an existing
			event with the same name for the same process, then
			this existing event is overwritten by the new one.		
			"""
			self.__scheduler.schedule_event_absolute (time, self, name)

	@typecheck
	def __init__ (self, eui64 = "0200000001020304", clock: optional (clock.Clock) = None, compression = False):
		super().__init__()
		self.__eui64 = store_data (eui64, Eui64Address)
		self.__port  = SixLowpanMessagePort (0, compression = compression)

		self.__scheduler = Scheduler (clock)

	# FIXME: function needed to avoid log errors
	def log_event (self, event):
		pass

	def get_scheduler (self):
		return self.__scheduler

	def get_message_port (self):
		return self.__port
	
	def reset (self):
		# clear all scheduled events
		self.__scheduler.reset()

		# generate the link local address (from the eui-64 id)
		self.__lladdr = IPv6Prefix("fe80::").make_address(self.__eui64)

		# by default we listen to two addresses:
		#  - our link-local address
		#  - the all-nodes multicast group ff02::1
		#
		# NOTE: the addresses are stored in a regular list (instead of
		# 	a set). Thus it is possible to have duplicates (which
		# 	is useful because we may register the same address with
		# 	multiple routers but we should not remove completely
		# 	the address when one router gets unreachable)
		self.__addresses = [self.__lladdr, IPv6Address("ff02::1")]


		# clear the message handlers
		self.__message_handlers = [] 
		
		# launch the processes
		self.__rdisc  = self.Process (self, self.__router_discovery_process)
		self.__icmpv6 = self.Process (self, self.__icmpv6_process)

		# dictionnary containing the process spawned for each router
		#	key:	IPv6 link local address of the router
		#	value:	register_process
		self.__router_processes = {}

		# blacklist contain all duplicated address
		self._black_list_addrDupl = []

		# blacklist contain all routeur with a full cache
		self._black_list_routeurFull = []


	def debug(self, msg):
		if self.config.debug:
			print (msg)
	
	def register_message (self, pattern, process):
		"""
		
		"""
		self.__message_handlers.append ((pattern, process))
	
	def run (self):
		"""Main loop
		
		The main loops monitors external events:
		 - messages received on the communication port
		 - scheduled event
		
		Then it reports the event to the adequate process.
		 - scheduled events are reported to the process that created
		   this event (by calling Process.schedule_event_xxxxxxx())
		 - messages are compared with the patterns registered (using
		   the register_message() function) and in case of a match, the
		   message is reported to the corresponding process (in case of
		   multiple matches, the message is reported multiple times)

		Events are reported to the process using the send() method
		(thus they can be received using the yield instruction).
		"""
		
		self.reset()

		while True:
			with alt:

				# match a scheduled event
				@self.__scheduler.event
				def _(ev):
					ev.process.send (ev.name)		

				# receive an IPv6 message from the communication port
				@self.__port.receive (IPv6())
				def _(value):
					# we get the IPv6 datagram
					ipv6_msg = value["pl"]["pl"]

					# filter out packets not addressed to this node
					if ipv6_msg["dst"] not in self.__addresses:
						print ("Addresses:", self.__addresses)
						print ("Ignored message (not for us):", ipv6_msg)
						return

					# check if one process is expecting this message
					# and report it if any
					for pattern, process in tuple(self.__message_handlers):
						if pattern.match (ipv6_msg):
							# report the IPv6 message to the process
							# (will be the result of the yield instruction)
							process.send (ipv6_msg)

				# receive any other message
				@self.__port.receive
				def _(value):
					print ("6lowpan implem: unexpected message:", value)
			
	def __send (self, msg):
		"""Send a message to the communication port

		This function chooses the destination (hw address) according to
		the IPv6 destination address and return True if successful.

		It will return False if the port is not connected or if there
		is no adequate route in the routing table.

		Note: it is not recommended to manipulate directly self.__port
		(eg: by calling self.__port.send()) because it may raise an
		exception if the port is not connected. This function will
		silently catch the exceptions.
		"""

		msg = msg.flatten()

		try:
			assert isinstance (msg, IPv6) # TODO: support 6lowpan & IPHC ?

			if msg["dst"][0] in (0xfe, 0xff): # FIXME: global multicast not supported
				# link-local or multicast destination
				self.__port.send (msg)
			else:
				# global destination
				# -> use the first known router
				if not self.__router_processes:
					# no route!
					return False

				self.__port.send (msg, dst = next(iter(self.__router_processes.keys())))

			return True

		except exceptions.PortNotConnected:
			return False


	def __retransTimer ( self, x ):
		if x < 3:
			return 10
		else:
			return min ( 2 * self.__retransTimer(x - 1) , 60 ) 


	def __send_rs (self, ipv6_destination_address):
		self.__send (
			IPv6 (src=self.__lladdr, dst=ipv6_destination_address) /
			ICMPv6RSol (opt =[ICMPv6SLL (hw = self.__eui64)]))

	def __send_ns (self, router_addr, prefix_info):

		# compute the lifetime to be used
		if not prefix_info.valid():
			lifetime = 0
		else:
			lifetime = prefix_info.choose_ns_lifetime()

			assert lifetime > 0

		self.__send (
			IPv6(src = prefix_info.address , dst = router_addr ) /
			ICMPv6NSol ( tgt=router_addr,opt = [
					ICMPv6SLL ( hw=self.__eui64 ) ,
					SixARO(eui= self.__eui64, lt = lifetime)
					]
				)
			)


##########################################################################################################################
###################################################### ROUTER DISCOVERY PROCESS ##############################################
#########################################################################################################################



	def __router_discovery_process (self, process):
		"""Router Discovery Process
		
		This process monitors the Router Advertisement received on the
		link so as to maintain the router list up-to-date. It spawns
		one register process for each router.

		If the router list is empty, then it will send and retransmit
		Router Solicitations.
		"""

		# Register the messages to be handled by this process (RA)	
		self.register_message (IPv6()/ICMPv6RAdv(), process)

		# dict containing the expiration time for each router
		router_expiry_times = {}

		# number of retransmissions done so far
		retransmission_count = 0

		# Schedule a event named "SendRS" to wake up when it is time to send a new Router Solicitation
		# (we delay the first RS so as not to send it before the beginning of the testcase)
		process.schedule_event_relative (self.config.DelayRS, "SendRSol")



		#It update the blacklist of router cache full
		def black_list_RouterFull_updated():
			"""This function shall be called each time one
			Router address is added to the blacklist.
			"""
			
			if self._black_list_routeurFull[len(self._black_list_routeurFull)-1] in router_expiry_times :
				router_expiry_times.pop(self._black_list_routeurFull[len(self._black_list_routeurFull)-1])



		while True:
			#-------------------- Schedule an expiration event if needed ------------------#
			if router_expiry_times:
				process.schedule_event_absolute (min (router_expiry_times.values()), "Expiry")

			print ("Disc: yield", router_expiry_times)
			msg = yield
			try:	
				#-------------------- Check for expired routers ------------------#
				now = self.__scheduler.time()

				print ("Disc:", now, msg)
				
				for rtr, expiry in tuple (router_expiry_times.items()):
					if expiry <= now:
						# kill the existing process
						print("Kill the existing process")
						rtr_process = self.__router_processes[rtr]

						kill_result = rtr_process.send("Kill")

						assert kill_result == "Finished"

						# remove it from the lists
						self.__router_processes.pop (rtr)
						router_expiry_times.pop (rtr)
						
						if not router_expiry_times:
							# the router list is now empty
							# -> restart sending multicast Router Solicitations
							retransmission_count = 0
							process.schedule_event_relative (0, "SendRSol")
					

				
				#-------------------- Event to send a Router Sollicitation ------------------#
				
				if msg == "SendRSol":
					# we only send a router solicitation when the router list is empty


					if not router_expiry_times:
						# We send the Router Sollicitation
						self.__send_rs ("ff02::2")

						retransmission_count += 1

						# schedule the retransmission of the next RS

						# --we must retransmit 3 packet during 10 seconds, then
						# --we SHOULD do binary exponential backoff of the retransmission for the others
						# timer for each subsequent retransmission
						if retransmission_count <= 2:
							delay = 10
						elif retransmission_count == 3:
							delay = 20
							
						elif retransmission_count == 4:
							delay = 40
						else:
							delay = 60

						process.schedule_event_relative (delay, "SendRSol")

						self.debug("RdiscProcess: RS sending scheduled in %d seconds." % delay)


				#-------------------- it is time to update full cache router list ------------#
				if msg == "blacklistRtrFullUpdated" :
					black_list_RouterFull_updated()

				

				#-------------------- When we receive a Router Advertisement ------------------#

				if (IPv6()/ICMPv6RAdv()).match (msg):
					# we received a Router Advertisement

					if not (IPv6()/ICMPv6RAdv(opt = Superset (ICMPv6OptionList , [ ICMPv6PI() ] ))).match(msg):
						print("The option ICMPv6PI is not contained in this Router Advertisement")
						continue

					router_addr = msg["src"]

					assert (router_addr in router_expiry_times) == (router_addr in self.__router_processes) # these two dict must always contain the same keys
					#We must check that the routeur address is not in the blacklist router
					addr_is_ok = True

					for router_full in self._black_list_routeurFull :
						if router_full == router_addr :
							addr_is_ok = False
							
					if addr_is_ok:	

						if router_addr not in router_expiry_times:
							# unknown router
							# -> start a new register process

							# spawn the new process
							p = self.Process (self, self.__router_registration_process, router_addr)
							self.__router_processes[router_addr] = p

							# send the RA message to the process
							p.send (msg)

						# compute the expiration time
						router_expiry_times[router_addr] = now + msg["pl"]["rlt"]

			except Exception as e:
				traceback.print_exc()




		
		



##########################################################################################################################
###################################################### REGISTRATION PROCESS ##############################################
##########################################################################################################################




	def __router_registration_process (self, process, router_addr):
		
		ra_template = IPv6(src = router_addr)/ICMPv6RAdv()
		na_template = IPv6(src = router_addr)/ICMPv6NAdv(opt = Superset(ICMPv6OptionList, [SixARO(eui=self.__eui64,len=2)]))

		self.register_message (ra_template, process)
		self.register_message (na_template, process)

		# expiration time of the router
		# (sys.maxsize if none)
		router_expiry_time = sys.maxsize

		# boolean used to ensure that we send at most one unicast RS in
		# case the router information need to be refreshed (FIXME: shall we retransmit it ?)
		unicast_rs_sent = True

		# list of prefix info classes
		#	key:	prefix
		#	value:	PrefixInfo
		prefix_infos = {}


		# list of context info classes
		#	key:	context
		#	value:	ContextInfo
		context_infos = {}

		# list of prefix infos for which we want to register an address
		register_queue = []
			

		# PrefixInfo being registered
		pending_registration = None

		
		#It update the blacklist of address duplicated
		def black_list_AddrDupl_updated():
			"""This function shall be called each time one
			address is added to the blacklist.
			"""

			# we stock the last element insert (maybe a little hard...)
			addr_dupl = self._black_list_addrDupl[len(self._black_list_addrDupl)-1]


			#we get the prefix in the prefix info list
			for pref in tuple ( prefix_infos.values()):
				if addr_dupl == pref.address :
					prefix_tosend = pref
				
			for blackaddr in self._black_list_addrDupl :
				if blackaddr==addr_dupl :
					#We affect now to expiry_time_valid to provoc a lifetime to 0
					prefix_tosend.expiry_time_valid=now
					self.__send_ns(router_addr,prefix_tosend)
					break


		def context_updated (mycontext):
				
			for cont in tuple (context_infos.values()):

				# if our context match in context_infos entry
				if mycontext.contextId == cont.contextId :
				
					if mycontext.valid():

						#Update of context values
						cont.c = mycontext.c
						cont.pf = mycontext.pf

						# Compute the new expiry time
						cont.refresh(lt,now)

					else :
						# The context is no valid

						# -> delete it
						context_infos.pop (cont.contextId)

			
			# schedule the next PrefixInfoUpdated event
			process.schedule_event_absolute (mycontext.expiry_time_valid, "ContextUpdated")

				

	




		def prefix_infos_updated ():
			"""This function shall be called each time one
			prefix information may have changed.
			"""

			wakeup_time = sys.maxsize

			for info in tuple (prefix_infos.values()):

				if info.should_register():
					# Valid prefix (that we should register)
					# or for which our registration is about to expire

					# schedule its registration
					if info not in register_queue:
						register_queue.append (info)

				if info.registration_expired():
					# Expired registration (registered address no longer valid)

					assert info.registered

					# deconfigure the address
					self.__addresses.remove (info.address)
					info.registered = False

					print ("Removing address %s" % info.address) 

				if not info.valid():
					# The prefix is no longer valid

					# -> delete it
					prefix_infos.pop (info.prefix)

				if info.registered:
					if info.ns_sent or info in register_queue:
						# next action is expiration
						my_wakeup_time = info.expiry_time_registration
						print ("my next expiration", my_wakeup_time)
					else:
						# next action is to renew the registration
						my_wakeup_time = info.expiry_time_registration - self.config.AddressRefreshTime
						print ("my next renew", my_wakeup_time)

					if my_wakeup_time < wakeup_time:
						wakeup_time = my_wakeup_time

			# schedule the next PrefixInfoUpdated event
			process.schedule_event_absolute (wakeup_time, "PrefixInfosUpdated")


		def schedule_unicast_rs():
			"""Compute the time at which we must wake up to send a
			new unicast RS to refresh the router information and
			schedule the event"""
			
			if not unicast_rs_sent:
				expiry_time = router_expiry_time

				for info in prefix_infos.values():
					expiry_time = min (expiry_time,
						info.expiry_time_valid,
						info.expiry_time_preferred)

				refresh_time = expiry_time - self.config.routerRefreshTime

				process.schedule_event_absolute (refresh_time, "SendRSol")

				print ("Will send a unicast RS to %s in %d seconds" % (str(router_addr), refresh_time))


		class ContextInfo:
			def __init__ (self, context,contextId,c,pf, lt, current_time):
				
				# IPv6 context
				self.context  = context
				self.contextId  = contextId
				self.c = c
				self.pf = pf
				self.lt = lt
				
				self.refresh (lt, current_time)
				#print("CONTEXT = '%s' c = '%d' clen= '%d' pf= '%s' lt='%d' " % (self.context 							self.contextId,self.c,self.clen,self.pf,self.lt))

			def refresh (self, lt, current_time):

				# Expiration time of the prefix (Valid Lifetime)
				self.expiry_time_valid = now + self.lt
				
				
			def valid (self):
				"""Return True of the context is still valid"""
				return now < self.expiry_time_valid

			


		class PrefixInfo:
			def __init__ (self, prefix, pi_option, current_time, hw_identifier):

				assert prefix == IPv6Prefix (pi_option["Prefix"], pi_option["PrefixLength"])

				# IPv6 prefix
				self.prefix  = prefix

				# IPv6 address generated for this prefix
				self.address = prefix.make_address (hw_identifier)

				# State variable indicating that this address
				# is registered with the router and configured
				# on this node
				self.registered  = False

				# State variable set to True when a NS has been
				# sent to renew the registration of the address
				self.ns_sent = False

				# Expiration time of the current registration
				self.expiry_time_registration = 0

				self.refresh (pi_option, current_time)

			def refresh (self, pi_option, current_time):

				assert self.prefix == IPv6Prefix (pi_option["Prefix"], pi_option["PrefixLength"])

				# Expiration time of the prefix (Valid Lifetime)
				self.expiry_time_valid     = now + pi_option["ValidLifetime"]
				#TODO: rename PreferedLifetime as PreferredLifetime

				# Expiration time of the prefix (Preferred Lifetime)
				self.expiry_time_preferred = now + pi_option["PreferedLifetime"]




			def valid (self):
				"""Return True of the prefix is still valid (according to the Valid Lifetime in the previous RA)"""
				return now < self.expiry_time_valid

			def preferred (self):
				"""Return True of the prefix is still preferred (according to the Preferred Lifetime in the previous RA)"""
				return now < self.expiry_time_preferred

			def should_register (info):
				"""Return true if we need to send a NS to register/extend the registration of the address"""

				#if the address is in the blacklist -> return false 
				for black_addr in self._black_list_addrDupl : 
					if black_addr==self.address :
						return false

				if not info.registered:
					return info.valid()

				else:
					return not info.ns_sent and ((info.expiry_time_registration - self.config.AddressRefreshTime ) <= now < info.expiry_time_registration)

			def registration_expired (self):
				"""Return true if the currently registered address has expired"""
				return self.registered and (now >= self.expiry_time_registration)

			def choose_ns_lifetime (info):
				"""Returns the desired registration lifetime in minutes (to be inserted in the NS).
				
				This function also stores the requested	expiration time internally
				in info.expiry_time_registration_requested.
				"""

				# TODO: ensure it is not less than the router lifetime
				lifetime = min (	info.expiry_time_valid     - now,
						info.expiry_time_preferred - now) // 60

				# minimum lifetime will be 1 minute
				if lifetime <= 0:
					lifetime = 1

				# remember the requested expiration time
				info.expiry_time_registration_requested = min (
					now + lifetime * 60,
					info.expiry_time_valid
				)

				print ("Requested lifetime for %s is %d min (expires at %d)" % (info.address, lifetime, info.expiry_time_registration_requested))

				return lifetime


		while True:
			print ("Reg: yield")
			msg = yield
			try:
				# TODO: wake-up to renew/revoke addresses
				now = self.__scheduler.time()
				
				print ("Reg:", now, msg)

				#-------------------- termination signal -------#

				if msg == "Kill":
					print ("Killing registration process for", router_addr)
					
					# deconfigure the registered addresses
					for info in prefix_infos.values():
						if info.registered:
							self.__addresses.remove (info.address)

					while True:
						yield "Finished"
				
				#-------------------- it is time to refresh the router informations -------#

				elif msg == "SendRSol":
					if not unicast_rs_sent:
						print( "send RS", router_addr)
						self.__send_rs (router_addr)
						unicast_rs_sent = True

				#-------------------- some actions are needed on some prefixes ------------#
				elif msg == "PrefixInfosUpdated":
					prefix_infos_updated()

				#-------------------- some actions are needed on the context ------------#
				elif msg == "ContextUpdated" :
					context_updated()

				#-------------------- it is time to update duplicated address list ------------#
				elif msg == "blacklistAddrDuplUpdated" :
					black_list_AddrDupl_updated()

				#-------------------- it is time to update full cache router list ------------#
				elif msg == "blacklistRtrFullUpdated" :
					black_list_RouterFull_updated()



				#-------------------- we received a Router Advertisement ------------------#

				elif (ra_template.match (msg)):
					# we received a Router Advertisement
					ra = msg["pl"]

					router_expiry_time = now + ra['rlt']

					# parse the Prefix Information options
					for pio in filter (lambda x: isinstance(x, ICMPv6PI), ra["opt"]):

						# Ignore PI options with the OnLink flag set
						if pio["l"] == 1:
							print("The option ICMPv6 has the bit l set to 1 !")
							continue

						prefix = IPv6Prefix (pio["Prefix"], pio["PrefixLength"])
						
						if prefix not in prefix_infos:
							# this is a new prefix!
							info = PrefixInfo (prefix, pio, now, self.__eui64)
							prefix_infos[prefix] = info
						else:
							# unknown prefix
							# -> refresh it
							info = prefix_infos[prefix]
							info.refresh (pio, now)

					#Context Informations

					if (IPv6()/ICMPv6RAdv(opt = Superset (ICMPv6OptionList,[ SixCO()] ))).match(msg) :
						
						
						context = msg["pl"]["opt"][SixCO]
						#ContextId is the key for context_infos
						contextId = msg["pl"]["opt"][SixCO]["cid"]
						compress = msg["pl"]["opt"][SixCO]["c"]
						contextPrefix = msg["pl"]["opt"][SixCO]["pf"]
						contextLt = msg["pl"]["opt"][SixCO]["lt"]
						

						if context not in context_infos:
							#this is a new context
							cont = ContextInfo(context,contextId,compress,contextPrefix,contextLt,now)
							context_infos[contextId] = cont
						else:
							# unknown context
							# -> refresh it
							cont = context_infos[contextId]
							info.refresh (lt, now)

							# Update information about the context
							context_updated(cont)
						


					# schedule next actions on prefix infos
					prefix_infos_updated()
					
					# schedule an event for sending a refresh unicast RS
					unicast_rs_sent = False
					schedule_unicast_rs()

				

				#-------------------- Address Registration ------------------#

				if pending_registration:
					finished = False
					

					if na_template.match(msg):
						# We received the Neighbor Advertisemen
						finished = True
						#FIXME: maybe check that the destination address is valid (which one shall be used ???)
						print("MATCH")
						aro = msg["pl"]["opt"][SixARO]
						
						if aro["Status"] == 0:
							# Successful registration !
							pending_registration.registered = True
							pending_registration.ns_sent = False

							# configure the ipv6 address on the node
							self.__addresses.append (pending_registration.address)
							print ("Successfully registered", pending_registration.address)
							print ("Addresses:", self.__addresses)

							prefix_infos_updated()
							
						elif aro["Status"] == 1:
						#A Status code of one indicates that the Address is duplicated
						# We should append it in the blacklist
							print("Status -> 1")
							self._black_list_addrDupl.append(pending_registration.address)

						#We send a message to all processes for an updating but not the current process
							
							for proc in tuple (self.__router_processes .values()):
									proc.schedule_event_relative (0,"blacklistAddrDuplUpdated")

							
						elif aro["Status"] == 2:
							#A Status code of two indicates that the Neighbor Cache of that router
  							#is full.  In this case the host SHOULD remove this router from its
 							#default router list and attempt to register with another router.  If
   							#the host has no more default routers it needs to revert to sending Router Solicitations

							# We should append it in the blacklist
							self._black_list_routeurFull.append(router_addr)
							#We kill the process
							process.schedule_event_relative (-10,"Kill")

							#We send a message to all processes for an updating of the blacklist Routeur
							self.__rdisc.schedule_event_relative (0,"blacklistRtrFullUpdated")
							

						else :
							print("Status !=0,1,2")

					elif msg == "TimeoutNS":
						# registration failed
						finished = True

					if finished:
						# registration terminated !

						try:
							register_queue.remove (pending_registration)
						except ValueError:
							pass
											
						pending_registration = None

				if not pending_registration: # NOT ELSE !!!
					
					# start a new registration !
					
					while register_queue:
						pending_registration = register_queue.pop()

						if pending_registration.should_register():

							assert pending_registration.valid()
							
							# Send a neighbour solicitation
							self.__send_ns (router_addr, pending_registration)

							pending_registration.ns_sent = True
							pending_registration.expiry_time_registration = pending_registration.expiry_time_registration_requested

							# schedule a 5 second timeout (if successful)
							process.schedule_event_relative (5, "TimeoutNS")

							break
					
			except Exception as e:
				traceback.print_exc()



##########################################################################################################################
###################################################### ICMPv6 PROCESS(Echo-reply)  ##############################################
##########################################################################################################################



	def __icmpv6_process (self, process):
		'''ICMPv6 process
		
		- answers echo requests sent to this node
		'''
		self.register_message (IPv6()/ICMPv6EchoRequest(), process)

		while True:
			msg = yield
			self.debug("ICMPv6Process: Got an ICMPv6 Echo Request From:"+ repr(msg["src"]) +" , to: " + repr(msg["dst"]) )
			
			
			if msg["dst"] not in self._black_list_addrDupl :

				if msg["dst"][0] in (0xfe, 0xff): # linklocal or multicast
					src = self.__lladdr	#FIXME: this covers only the link-local scope
				else:
					src = msg["dst"]

				ereq = msg["pl"]

	
				self.__send (
					IPv6 ( src = src, dst = msg["src"])/
					ICMPv6EchoReply ( id = ereq["id"], seq = ereq["seq"], pl = ereq["pl"] )
				)



#!/usr/bin/env python3
#
#  (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re
from collections import OrderedDict
from .templates import *
from ttproto.core.analyzer import TestCase, is_protocol, Node, Conversation, Capture
from ttproto.core.dissector import Frame
from ttproto.core.templates import *
from ttproto.core.typecheck import typecheck
from ttproto.core.lib.all import *
from urllib import parse
from ttproto.core.exceptions import Error

# CoAP constants
RESPONSE_TIMEOUT = 2
RESPONSE_RANDOM_FACTOR = 1.5
MAX_RETRANSMIT = 4
MAX_TIMEOUT = 10 + round(
    (RESPONSE_TIMEOUT * RESPONSE_RANDOM_FACTOR) * 2 ** MAX_RETRANSMIT
)


class NoStimuliFoundForTestcase(Error):
    """
    Error raised when no stimuli was defined for the testcase
    """
    pass


class FilterError(Error):
    """
    The error thrown when an error occurs into Filter object during its run
    """
    pass


class CoAPTestCase(TestCase):
    """
    The test case extension representing a CoAP test case
    """

    # Some default parameters
    reverse_proxy = False
    urifilter = False

    @classmethod
    @typecheck
    def get_protocol(cls) -> is_protocol:
        """
        Get the protocol corresponding to this test case. This has to be
        implemented into the protocol's common test case class.

        :return: The protocol on which this TC will occur
        :rtype: Value
        """
        return CoAP

    @classmethod
    @typecheck
    def get_nodes_identification_templates(cls) -> list_of(Node):
        """
        Get the nodes of CoAP test case.

        :return: The nodes involved in the CoAP test case
        :rtype: [Node]

        .. note:: For CoAP it is simple so we can define general behaviour for all CoAP test cases,
                  for other protocols this may have to be defined in subclasses (specific test case scenario).
        """
        return [
            Node('client', UDP(dport=5683)),
            Node('server', UDP(sport=5683))
        ]

    @classmethod
    @typecheck
    def preprocess(
            cls,
            capture: Capture,
            expected_frames_pattern: list_of(Value)
    ) -> (list_of(Conversation), list_of(Frame)):
        """
        Pre-processes and filters the frames of the capture into test case related
        conversations.

        This method depends on the protocol features, so it cannot be
        implemented in a generic way (in super class TestCase)

        :param Capture: The capture which will be filtered/preprocessed
        :return: list of conversations and list of ignored frames
        """

        if not expected_frames_pattern:  # If there is no stimuli at all
            raise NoStimuliFoundForTestcase(
                'Expected stimuli declaration from the test case for running pre-process and filtering of frames'
            )
        conversations_created_by_token, ignored = cls.extract_all_coap_conversations(capture)
        conversations_correlated_for_testcases = cls.correlate(conversations_created_by_token, expected_frames_pattern)
        return conversations_correlated_for_testcases, ignored

    @classmethod
    @typecheck
    def correlate(cls, conversations: list_of(Conversation),
                  expected_frames_pattern: list_of(Value)) -> list_of(Conversation):
        """
        Correlates related conversations.
        Conversations related to a test case, having several
        stimulis, are merged into one.
        This method relies on the given stimulis as frames pattern to determine which
        conversations are related for the test case.

        When having several stimulis, we rely on timestamp to have the wanted order.
        A whole related conversation will be fully inserted BEFORE
        the first frame of the base conversation having a timestamp
        greater than the timestamp of the related one.

        Simple example with only one stimulis :

        Here is a simple example with a single stimulis for TD_COAP_CORE_04.
        The stimulis for this TC is a confirmable POST request on resource /test
        with empty or non-empty payload.
        I.e CoAP(type='con', code='post', opt=Opt(CoAPOptionUriPath("test")).

        Let's say we have a 4 conversations in the given list of conversations:

        # python3 -i console.py
        >>> from pprint import pprint
        >>> c = Capture('tests/test_dumps/preprocess/coap/Two_tc_two_times_each_with_overlap.pcap')
        >>> pprint(c.frames)
        [<Frame   1: [127.0.0.1 -> 127.0.0.1] CoAP [CON mid 19207] GET /separate, tok 00 00 5f 6a >,
         <Frame   2: [127.0.0.1 -> 127.0.0.1] CoAP [ACK mid 19207] Empty >,
         <Frame   3: [127.0.0.1 -> 127.0.0.1] CoAP [CON mid 56421] 2.05 Content , tok 00 00 5f 6a >,
         <Frame   4: [127.0.0.1 -> 127.0.0.1] CoAP [ACK mid 56421] Empty >,
         <Frame   5: [127.0.0.1 -> 127.0.0.1] CoAP [CON mid 4146] POST /test, tok 28 6b 05 55 ca 09 29 ae >,
         <Frame   6: [127.0.0.1 -> 127.0.0.1] CoAP [ACK mid 4146] 2.01 Created /location1/location2/location3, tok 28 6b 05 55 ca 09 29 ae >,
         <Frame   7: [127.0.0.1 -> 127.0.0.1] CoAP [CON mid 10512] GET /separate, tok 00 00 5b e1 >,
         <Frame   8: [127.0.0.1 -> 127.0.0.1] CoAP [ACK mid 10512] Empty >,
         <Frame   9: [127.0.0.1 -> 127.0.0.1] CoAP [CON mid 58373] POST /test, tok e4 d8 1e c4 96 6e f9 20 >,
         <Frame  10: [127.0.0.1 -> 127.0.0.1] CoAP [ACK mid 58373] 2.01 Created /location1/location2/location3, tok e4 d8 1e c4 96 6e f9 20 >,
         <Frame  11: [127.0.0.1 -> 127.0.0.1] CoAP [CON mid 56422] 2.05 Content , tok 00 00 5b e1 >,
         <Frame  12: [127.0.0.1 -> 127.0.0.1] CoAP [ACK mid 56422] Empty >]



        conversations[0] = [
            < [Client 1 -> Server] CoAP [CON 19207] GET /separate >
            < [Server -> Client 1] CoAP [ACK 19207] Empty >
            < [Server -> Client 1] CoAP [CON 56421] 2.05 Content >
            < [Client 1 -> Server] CoAP [ACK 56421] Empty >
        ]

        conversations[1] = [
            < [Client 1 -> Server] CoAP [CON 4146] POST /test>
            < [Server -> Client 1] CoAP [ACK 4146] 2.01 Created>
        ]

        conversations[2] = [
            < [Client 1 -> Server] CoAP [CON 10512] GET /separate>
            < [Server -> Client 1] CoAP [ACK 10512] Empty >
            < [Server -> Client 1] CoAP [CON 56422] 2.05 Content >
            < [Client 1 -> Server] CoAP [ACK 56422] Empty >
        ]

        conversations[3] = [
            < [Client 1 -> Server] CoAP [CON 58373] POST /test>
            < [Server -> Client 1] CoAP [ACK 58373] 2.01 Created>
        ]

        From a list of those conversations and the stimulis of TD_COAP_CORE_04
        as argument, this method will return a new list of conversations
        containing only the second (i.e conversations[1]) and the fourth
        conversations (conversations[3]) because only those correspond to the
        stimulis.

        The two others conversations are ignored.
        Of course if the expected_frames_pattern argument would be the
        stimulis of TD_COAP_CORE_09 instead of CORE_04, we would return a list
        contaings conversations[0] and conversations[2] because those matches
        with the stimulis of TD_COAP_CORE_09.
        This stimulis is a GET request on /separate,
        i.e CoAP(type='con', code='get', opt=Opt(CoAPOptionUriPath("separate")).

        If we would pass both stimulis of TD_COAP_CORE_09 and TD_COAP_CORE_04,
        all conversations would be returned.

        Simple example with several stimulis :

        Here is a simple example with a single stimulis for TD_COAP_CORE_07.
        The first stimulis for this TC is a client sending a GET request with
        observe option to receive notification from the server, on resource /obs.

        The second stimulis is another CoAP client doing a DELETE request on the
        same resource.

        Let's say we extracted 5 conversations in the given pcap.

        conversations[0] = [
            <Frame   1: [Client 1 -> Server] CoAP [CON 45525] GET /obs>
            <Frame   2: [Server -> Client 1] CoAP [ACK 45525] 2.05 Content >
            <Frame   7: [Server -> Client 1] CoAP [CON 9255] 2.05 Content >
            <Frame   8: [Client 1 -> Server] CoAP [ACK 9255] Empty >
            <Frame  10: [Server -> Client 1] CoAP [CON 9256] 4.04 Not Found >
            <Frame  11: [Client 1 -> Server] CoAP [ACK 9256] Empty >
        ]

        conversations[1] = [
            <Frame   3: [Client 1 -> Server] CoAP [CON 10074] POST /test>
            <Frame   4: [Server -> Client 1] CoAP [ACK 10074]
                                    2.01 Created /location1/location2/location3>
        ]

        conversations[2] = [
            <Frame   5: [Client 1 -> Server] CoAP [CON 15501] GET /test>
            <Frame   6: [Server -> Client 1] CoAP [ACK 15501] 2.05 Content >
        ]

        conversations[3] = [
            <Frame   9: [Client 1 -> Server] CoAP [CON 51892] DELETE /obs>
            <Frame  12: [Server -> Client 1] CoAP [ACK 51892] 2.02 Deleted >
        ]

        conversation[4] = [
            <Frame  13: [Client 1 -> Server] CoAP [CON 65335] GET /test>
            <Frame  14: [Server -> Client 1] CoAP [ACK 65335] 2.05 Content >
        ]

        The first conversations (conversations[0]) match the first stimuli.
        The fourth conversation (conversations[3]) match the second stimuli.

        The other conversation don't match any stimulis, hence are discarded if
        the given stimulis are the stimulis is TD_COAP_OBS_07.

        As those two conversations above belong to the same instance
        of the same TC, they will be merged into a single conversation, that is:

        <Frame   1: [Client 1 -> Server] CoAP [CON 45525] GET /obs>
        <Frame   2: [Server -> Client 1] CoAP [ACK 45525] 2.05 Content >
        <Frame   7: [Server -> Client 1] CoAP [CON 9255] 2.05 Content >
        <Frame   8: [Client 1 -> Server] CoAP [ACK 9255] Empty >
        <Frame   9: [Client 2 -> Server] CoAP [CON 51892] DELETE /obs>
        <Frame  12: [Server -> Client 2] CoAP [ACK 51892] 2.02 Deleted >
        <Frame  10: [Server -> Client 1] CoAP [CON 9256] 4.04 Not Found >
        <Frame  11: [Client 1 -> Server] CoAP [ACK 9256] Empty >

        This method return a list_of(Conversation) because a futur feature that
        will allow to execute tests from a large pcap having several TC,
        with single TC able to being executed several times.
        Each different instance of the TC will have a different Conversation.

        """
        # TODO Adding an example in the documentation above.
        conversations_matching_stimulis = cls.__get_all_matching_conversations(
            conversations,
            expected_frames_pattern
        )
        conversations_to_merge = cls.__get_conversation_to_merge(
            conversations_matching_stimulis,
            expected_frames_pattern
        )
        return cls.__merge_all_conversations_to_merge(conversations_to_merge)

    @classmethod
    @typecheck
    def __merge_all_conversations_to_merge(
            cls,
            conversations_to_merge: list_of(list_of(Conversation))
    ) -> list_of(Conversation):
        """
        A list of list of conversation that must be merged together.
        That is, we merge all conversations of in the same "inner list" into one,
        and return the list of those merged conversations.
        """
        all_merged_conversations = []

        for list_of_related_conversations in conversations_to_merge:
            merged_conversation = cls.__merge_conversations(
                list_of_related_conversations)
            all_merged_conversations.append(merged_conversation)

        return all_merged_conversations

    @classmethod
    @typecheck
    def __merge_conversations(
            cls,
            conversations_to_merge: list_of(Conversation)
    ) -> Conversation:
        """
        Merge the given conversations into a single one.
        The given conversation should be belong to the same instance of the same TC.
        """
        nodes = cls.get_nodes_identification_templates()
        merged_conversation = Conversation(nodes)

        # The list of conversations already merged,
        # used to avoid processing two time the same one several times.
        added_in_merged_conv = list()

        for conv in conversations_to_merge:
            if conv in added_in_merged_conv:
                continue  # Already treated
            for frame in conv:
                for other_conv in conversations_to_merge:
                    if other_conv is not conv \
                            and other_conv[0].timestamp < frame.timestamp \
                            and other_conv not in added_in_merged_conv:
                        # We check at the first frame of other conversation to see
                        # if there is any correlation.
                        for other_frame in other_conv:
                            merged_conversation.append(other_frame)
                            added_in_merged_conv.append(other_conv)
                merged_conversation.append(frame)
            added_in_merged_conv.append(conv)

        return merged_conversation

    @classmethod
    @typecheck
    def __get_all_matching_conversations(
            cls,
            conversations: list_of(Conversation),
            expected_frames_pattern: list_of(Value)
    ) -> list_of(Conversation):
        """

        Retrieve a list of all conversations that correspond to any stimulis
        of the test cases (see get_stimulis()).

        """
        conversations_to_merge = list()

        for current_conversation in conversations:
            for frame in current_conversation:
                # conv_already_added is here to avoid doing useless iteration
                # and adding several times the same conv.
                conv_already_added = False
                for stimulis in expected_frames_pattern:
                    if frame[CoAP] in stimulis:
                        # We only look at the first frame as stimulis are
                        # Always query from client.
                        conversations_to_merge.append(current_conversation)
                        conv_already_added = True
                        break
                if conv_already_added:
                    break

        return conversations_to_merge

    @classmethod
    @typecheck
    def __get_conversation_to_merge(
            cls,
            conversations: list_of(Conversation),
            expected_frames_pattern: list_of(Value)) \
            -> list_of(list_of(Conversation)):
        """
        Retrieve a list 'outer' of list 'inter' with each conversations inside
        those inter lists having to be merged together. That is, conversations
        must be merged together if they belong to the same instance of the same
        TC described with expected_frames_pattern.

        This method return a list of list and not a simple list,
        because the user may execute several times the same test.
        Also, if running passive test from a pcap, this pcap may contains
        several conversations related to the same TC.
        """
        # TODO example
        conversations_maching_a_stimulis = conversations
        all_convs_to_merge = []
        current_convs_to_merge = []

        for conv in conversations_maching_a_stimulis:
            if conv[0][CoAP] in expected_frames_pattern[0] \
                    and len(current_convs_to_merge) == 0:
                current_convs_to_merge.append(conv)
            elif conv[0][CoAP] in expected_frames_pattern[0]:
                all_convs_to_merge.append(current_convs_to_merge)
                current_convs_to_merge = []
                current_convs_to_merge.append(conv)
            else:
                current_convs_to_merge.append(conv)

        if current_convs_to_merge not in all_convs_to_merge:
            all_convs_to_merge.append(current_convs_to_merge)

        return all_convs_to_merge

    @classmethod
    @typecheck
    def extract_all_coap_conversations(
            cls,
            capture: Capture) -> (list_of(Conversation), list_of(Frame)):

        protocol = cls.get_protocol()
        nodes = cls.get_nodes_identification_templates()

        conversations = []
        ignored = []

        # TODO what happens if no protocol declared on the test case?
        if not nodes or len(nodes) < 2:
            raise ValueError(
                'Expected at leaset two nodes declaration from the test case'
            )
        if not protocol:
            raise ValueError(
                'Expected a protocol under test declaration from the test case'
            )
        # Get the frames related with the protocol under test
        frames, ignored = Frame.filter_frames(capture.frames, protocol)

        # Map a token to the corresponding conversations.
        tkn_to_conv = OrderedDict()

        # Map a MID of a frame to it's containing token.
        # It's allow us to know to which conversation ACK frames belongs to
        # even though ACK do not contains token.
        mid_to_tkn = OrderedDict()

        # A set of CMID of messages for which we already saw an ACK.
        # This set allows use to detect duplicated ACK.
        acknowledged_CMID = set()

        for frame in frames:
            CMID = frame[CoAP]["mid"]
            CTOK = frame[CoAP]["tok"]

            if frame[CoAP]["type"] == 2 or frame[CoAP]["type"] == 3:
                if CMID in acknowledged_CMID:
                    # A duplicated ACK or RST
                    ignored.append(frame)
                    continue
                else:
                    acknowledged_CMID.add(CMID)
                    if CMID not in mid_to_tkn:
                        # An orphan ACK or RST
                        ignored.append(frame)
                    else:
                        tkn_to_conv[mid_to_tkn[CMID]].append(frame)
            else:
                if CTOK in tkn_to_conv:
                    tkn_to_conv[CTOK].append(frame)
                else:  # First time we encounter the token "CTOK".
                    conv = Conversation(nodes)
                    conv.append(frame)
                    tkn_to_conv[CTOK] = conv
                mid_to_tkn[CMID] = CTOK

        conversations = [tkn_to_conv[k] for k in tkn_to_conv.keys()]
        return conversations, ignored

    @typecheck
    def next_skip_ack(self, optional: bool = False):
        """
        Call self.next() but skips possibly interleaved ACKs

        :param optional: If we have to get a next frame or not
        :type optional: bool
        """

        # Goes to next frame
        self.next(optional)

        # While there is one and that it's an ack, pass it
        while all((
                    self._frame is not None,
                    self._frame[CoAP] in CoAP(type='ack', code=0)
        )):
            self.next(optional)

    @property
    def coap(self) -> CoAP:
        """
        Get the coap layer of the current frame

        :return: The coap layer of the current frame
        :rtype: CoAP
        """
        return self._frame[CoAP]

    @typecheck
    def uri(self, uri: str, *other_opts):
        """
        Filter for disabling a template if URI-Filter is disabled

        :param uri: The uri
        :param other_opts: More options
            Elemements may be either:
                - CoAPOption datas
                    -> will be fed into a Opt() together with the Uri options
                - CoAPOptionList datas
                    -> will be combined with the Opt() within a All() template
        :type uri: str
        :type other_opts: tuple of more parameters
        """
        opt = []
        opt_list = []
        for o in other_opts:
            if issubclass(o.get_type(), CoAPOption):
                opt.append(o)
            elif issubclass(o.get_type(), CoAPOptionList):
                opt_list.append(o)
            else:
                raise ValueError

        if self.urifilter:
            u = urllib.parse.urlparse(uri)
            if u.path:
                assert not any(
                    isinstance(v, CoAPOptionUriPath) for v in other_opts
                )
                for elem in u.path.split("/"):
                    if elem:
                        opt.append(CoAPOptionUriPath(elem))
            if u.query:
                assert not any(
                    isinstance(v, CoAPOptionUriQuery) for v in other_opts
                )
                for elem in u.query.split("&"):
                    if elem:
                        opt.append(CoAPOptionUriQuery(elem))

        if opt:
            opt_list.append(Opt(*opt))

        if not opt_list:
            return None
        elif len(opt_list) == 1:
            return opt_list[0]
        else:
            return All(*opt_list)


class Link(list):
    """
    Class representing the link values for CoAP

    .. example:: coap://[adress][/uri]?[par_name]=[par_token]
    .. note:: In the whole class, the string representation of the link value
              got by parsing the payload of the CoAP packet is stored into the
              's' variable
    """

    __re_uri = re.compile(r"<([^>]*)>")
    __re_par_name = re.compile(r";([0-9A-Za-z!#$%&+^_`{}~-]+)(=?)")
    __re_ptoken = re.compile(r"[]!#$%&'()*+./:<=>?@[^_`{|}~0-9A-Za-z-]+")

    class FormatError(Exception):
        """
        Error thrown when there is a format error during parsing the uri value
        """
        pass

    @typecheck
    def is_compiled_regex(arg):
        """
        Check if a parameter is a valid regex compiled object.
        This function is used for the typechecker decorator.

        :return: True if a valid compiled regex, False if not
        :rtype: bool
        """
        return all((
            arg is not None,
            isinstance(arg, type(re.compile('dummy_pattern')))
        ))

    @typecheck
    def __init__(self, pl: bytes):
        """
        Initialize the link representation

        :param pl: The payload of the CoAP layer
        :type pl: bytes
        """

        @typecheck
        def error(msg: str, in_string=None):
            """
            Function to manage format errors

            :param msg: The string that we were seeking
            :param in_string: The object in which we were seeking
            :type msg: str
            :type in_string: anything

            :raises FormatError: Always raises this exception because it's the
                                 main purpose of this function
            """
            if in_string:
                raise self.FormatError("%s in %r" % (msg, in_string))
            else:
                raise self.FormatError("%s at %r..." % (msg, s[:40]))

        @typecheck
        def have(string: str) -> bool:
            """
            Check if the link value contains the entered parameter

            :param string: The string to seek in the link value
            :type string: str

            :return: True if the string is contained into the link value
            :rtype: bool
            """
            return s and s.startswith(string)

        @typecheck
        def percent_unquote(string: str) -> str:
            """
            Replace '%xx' url encoded special characters by their string value

            :param string: The uri value from which we will replace those
            :type string: str

            :raises FormatError: If the urllib was unable to replace the '%xx'
                                 characters

            :return: The uri with the special characters as normal char
            :rtype: str
            """
            try:
                return urllib.parse.unquote(string, errors='strict')
            except UnicodeDecodeError as e:
                error(str(e), string)

        # Try to parse the given payload into utf-8 string representation
        try:
            s = str(pl, 'utf-8')
        except UnicodeDecodeError as e:
            error(str(e), pl)

        # Match object used here
        mo = None

        @typecheck
        def consume(pattern: either(str, is_compiled_regex), subject: str):
            """
            Consume a part of the link value (stored into s variable)

            :param pattern: The pattern to consume
            :param subject: The name of the subject to consume
            :type pattern: either(str, _sre.SRE_Pattern)
            :type subject: str
            """

            # Get the extern s (for uri current value) and mo (match object)
            nonlocal s, mo

            # If the pattern is a string
            if isinstance(pattern, str):
                if s.startswith(pattern):
                    s = s[len(pattern):]
                    return

            # If the pattern is a compiled regex
            else:
                mo = re.match(pattern, s)
                if mo:
                    s = s[mo.end(0):]
                    return

            # If it didn't match with the current uri value
            error("malformed %s" % subject)

        # If it managed to get a string (uri value) from the payload
        if s:
            while True:

                # Link-value
                consume(self.__re_uri, 'uri')

                # Get the uri value without the encoded special chars
                uri = percent_unquote(mo.group(1))

                # Store the value of it
                link_value = self.LinkValue(uri)

                # While there is a ";" we can process the parameter
                while have(";"):

                    # Link-param
                    consume(self.__re_par_name, 'parmname')

                    # Get the name of the parameter
                    name = mo.group(1)

                    # If no value associated to it
                    if not mo.group(2):
                        value = None

                    # If it does have a value associated
                    elif have('"'):

                        # Quoted-string => Read and unquote it
                        value = []
                        esc = False

                        # Parse the whole uri left
                        for i in range(1, len(s)):

                            # Get each character
                            c = s[i]

                            # If it's not escaped
                            if not esc:

                                # If it's escaped now
                                if c == '\\':
                                    esc = True

                                # End of string
                                elif c == '"':
                                    break

                                # Other chars
                                # TODO: Normalise LWS
                                else:
                                    value.append(c)

                            # If it is escaped
                            else:
                                esc = False

                                # Quoted char
                                if c == '"' or c == '\\':
                                    value.append(c)

                                # Was an unquoted \
                                else:
                                    value.append('\\' + c)

                        # Error: unterminated quoted-string
                        else:
                            error(
                                "attribute value for %r is %s"
                                %
                                (
                                    name,
                                    'an unterminated quoted-string'
                                )
                            )

                        # Transform the value char list into a single string
                        value = ''.join(value)

                        # If still empty
                        if not value:
                            error("attribute value for %r is empty" % name)

                        # Consume the read part
                        s = s[i + 1:]

                    # If it doesn't begin with a quote, it's a token
                    else:

                        # Consume it
                        consume(self.__re_ptoken, 'ptoken')

                        # And generate its value
                        value = percent_unquote(mo.group(0))

                    # Add the link value pair
                    link_value.append((name, value))

                # In the end, append the link value to the link list
                self.append(link_value)

                # If we finished reading the whole uri, s is now empty
                if not s:
                    break

                # Next link-value
                consume(",", "delimiter, expected ','")

    class LinkValue(list):
        """
        A class representing the link values which consist into a parameter
        name associated a parameter value
        """

        @typecheck
        def __init__(self, uri: str):
            """
            Initialize the link value object

            :param uri: The basic uri
            :type uri: str
            """
            self.uri = uri

        @typecheck
        def get(
                self,
                par_name: str,
                testcase: optional(CoAPTestCase) = None
        ) -> optional(str):
            """
            Get the value of a link value from its parameter name

            :param par_name: The parameter name
            :param testcase: The TestCase object to put its verdict to 'fail'
            :type par_name: str
            :type testcase: optional(CoAPTestCase)

            :return: The parameter value associated to this parameter name if
                     one found, None if none found
            :rtype: optional(str)
            """

            # The result to return, None at the beginning to check if a link
            # value has multiple values for a single parameter name
            result = None

            # For each couple (name => value) inside this list
            for name, value in self:

                # If the parameter name is found
                if name == par_name:

                    # If no result found until here, ok and put it
                    if result is None:
                        result = value

                    # If a value was already put and another one if found
                    else:
                        msg = (
                            "link-value contains multiple %r parameters"
                            %
                            par_name
                        )

                        # If a test case object is given, its verdict fails
                        if testcase:
                            testcase.set_verdict('fail', msg)
                        else:
                            raise Exception(msg)

            # Return the value if one found, if none found just return None
            return result


if __name__ == "__main__":
    pass



from ..common import *


class TD_COAP_CORE_12(CoAPTestCase):
    """
---
TD_COAP_CORE_12:
    cfg: CoAP_CFG_BASIC
    not: Not all clients may be able to send a zero-length Token
    obj: Perform GET transaction using empty Token (CON mode)
    pre: Server offers the resource /test with resource content is not empty
        that handles GET with an arbitrary payload
    ref: '[COAP] 2.2, 5.8.1, 5.10.1'
    seq:
    -   s: "Client is requested to send a confirmable GET request using\
            \ zero-length Token to server\u2019s resource "
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON)
            - Code = 1 (GET)
            - "Zero-Length Token \u2794 CTOK"
            - Uri-Path option "test"
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.05 (Content)
            - Message ID = CMID, Token = CTOK
            - Content-format option
            - Non-empty Payload
    -   v: Client displays the response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [
            CoAP(type='con', code='get', tok=b'', opt=Opt(CoAPOptionUriPath("test")))
        ]

    def run(self):
        self.match("client", CoAP(type="con", code="get", tok=b"", opt=self.uri("/test")))

        self.next_skip_ack()

        if self.match("server", CoAP(code=2.05, pl=Not(b""), opt=Opt(CoAPOptionContentFormat()), )):
            self.match("server", CoAP(tok=b""), "fail")



from ..common import *


class TD_COAP_CORE_07(CoAPTestCase):
    """
---
TD_COAP_CORE_07:
    cfg: CoAP_CFG_BASIC
    obj: Perform PUT transaction (NON mode)
    pre: Server offers a /test resource that handles PUT
    ref: '[COAP] 5.8.3, 5.2.3'
    seq:
    -   s:
        - 'Client is requested to send a PUT request with:'
        -   - Type = 1 (NON)
            - Code = 3 (PUT)
            - An arbitrary payload
            - Content-format option
    -   c:
        - 'The request sent by the client contains:'
        -   - Type=1 and Code=3
            - "Client-generated Message ID (\u2794 CMID)"
            - "Client-generated Token (\u2794 CTOK)"
            - Uri-Path option "test"
    -   v: Server displays the received information
    -   c:
        - 'Server sends response containing:'
        -   - Type = 1 (NON)
            - Code = 2.04 (Changed) or 2.01 (Created)
            - "Server-generated Message ID (\u2794 SMID)"
            - Token = CTOK
            - Content-format option if payload non-empty
            - Empty or non-empty Payload
    -   v: Client displays the received response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        # TODO client message w/ no content format and payload -> fail?
        return [
            CoAP(
                type='non',
                code='put',
                pl=Not(b''),
                opt=Opt(CoAPOptionContentFormat(), CoAPOptionUriPath("test"))
            )
        ]

    def run(self):
        if self._frame is None: #inconclusive instead of error is no frame at all.
            return
        self.match("client", CoAP(type="non", code="put", opt=self.uri("/test")))
        self.match("client", CoAP(pl=Not(b''), opt=Opt(CoAPOptionContentFormat()), ), "inconclusive")

        CTOK = self.coap["tok"]

        self.next()

        self.match("server", CoAP(type="non", code=Any(65, 68), tok=CTOK))

        if self.match('server', CoAP(pl=Not(b'')), None):
            self.match("server", CoAP(opt=Opt(CoAPOptionContentFormat())), "fail")



#!/usr/bin/env python3
#
#   (c) 2018  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#


from ttproto.core.lib.ieee802154 import Ieee802154
from ttproto.core.lib.inet.meta import *
from ttproto.core.lib.inet.basics import *
from ttproto.core.lib.inet.sixlowpan import SixLowpan

import ttproto.core.lib.inet.udp

__all__ = [
    'ZigBeeEncapsulationProtocol'
]

class ZigBeeEncapsulationProtocol(
    metaclass=InetPacketClass,
    fields=[
        ("Protocol ID string", "pid", Hex(UInt16)),
        ("Version", "ver", UInt8),
        ("Type", "type", UInt8),
        ("Channel", "ch", UInt8),
        ("Device ID", "did", UInt16),
        ("LQI/CRC Mode", "mode", UInt8),
        ("LQI Value", "lqi", UInt8),
        ("NTP timestamp (s)", "nts", Hex(UInt32)),
        ("NTP timestamp fraction (ms)", "nts_frac", Hex(UInt32)),
        ("Sequence number", "seq", UInt32),
        ("future", "fut", Hex(UInt80)),
        ("Length", "len", UInt8),
        ("Payload", "pl", Ieee802154),
    ]):
    pass

ttproto.core.lib.inet.udp.udp_port_map[17754] = ZigBeeEncapsulationProtocol



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from    ttproto.core.data import Value, BidictValueType
from    ttproto.core.lib.inet.meta import *
from    ttproto.core.lib.inet.basics import *
from    ttproto.core.lib.inet.ip import *
from    ttproto.core.lib.encap import encap_type_bidict
import ttproto.core.lib.ethernet

__all__ = [
    "IPv6",
    "IPV6_ALL_NODES",
    "IPV6_ALL_ROUTERS",
    "IPV6_UNSPECIFIED_ADDRESS",
]


class IPv6(
        metaclass=InetPacketClass,
        fields=[
            ("Version", "ver", UInt4, 6),
            ("TrafficClass", "tc", Hex(UInt8), 0),
            ("FlowLabel", "fl", Hex(UInt20), 0),
            ("PayloadLength", "len", UInt16, InetLength("Payload")),
            ("NextHeader", "nh", UInt8, InetType(ip_next_header_bidict, "Payload")),
            ("HopLimit", "hl", UInt8, 64),  # FIXME: should be 255 for NS messages
            ("SourceAddress", "src", IPv6Address, "::"),
            ("DestinationAddress", "dst", IPv6Address, "::"),
            ("Payload", "pl", Value, b"")
        ],
        descriptions={"nh": ip_next_header_descriptions},
):
    def describe(self, desc):
        desc.src = self["src"]
        desc.dst = self["dst"]
        if not self.describe_payload(desc):
            txt_desc = self.get_description("nh")
            desc.info = txt_desc if txt_desc else  "IPv6 nh: %d" % self["nh"]

        return True

    _build_message = InetPacketValue._build_message_ipv6_header


# tell the ethernet module that ether payload type 0x86dd should be mapped to the IPv6 class
ttproto.core.lib.ethernet.ethernet_type_bidict.update({
    0x86dd: IPv6,
})

# tell the encap module that encap  payload type 24, 28 and 30 should be mapped to the IPv6 class
ttproto.core.lib.encap.encap_type_bidict.update({
    24: IPv6,
    28: IPv6,
    30: IPv6
})

# map ip next header 41 to the IPv6 class (IPv6 over IPv6)
ip_next_header_bidict.update({
    41: IPv6,
})

#
# Well know Address
#

# TODO: implement a description dict of well known ipv6 addresses

# http://www.iana.org/assignments/ipv6-multicast-addresses/ipv6-multicast-addresses.xml
IPV6_ALL_NODES = IPv6Address("ff02::1")
IPV6_ALL_ROUTERS = IPv6Address("ff02::2")
IPV6_UNSPECIFIED_ADDRESS = IPv6Address("::")
